import sys
from unittest import mock

import numpy as np
import pytest
from hypothesis import given
from hypothesis import strategies as st

import tesliper.glassware.array_base as ab
from tesliper.exceptions import InconsistentDataError


@pytest.mark.parametrize(
    "values,lengths",
    [
        ([], ()),
        ([1], ()),
        ([[]], (0,)),
        ([[1]], (1,)),
        ([[1, 2], [1]], (2,)),
        ([[[1], [2]], [[1]]], (2, 1)),
        ([[[1, 2], [2]], [[1]]], (2, 2)),
        ([[[1], [2]], [[1, 2]]], (2, 2)),
        ([[[1], 2], [[1, 2]]], (2,)),
        ([[[1, 2]], [[1], 2]], (2,)),
    ],
)
def test_longest_subsequences(values, lengths):
    assert ab.longest_subsequences(values) == lengths


@pytest.mark.parametrize(
    "values,lengths", [(["a"], ()), ([["a"]], (1,)), ([["aa"]], (1,))]
)
def test_longest_subsequences_str(values, lengths):
    assert ab.longest_subsequences(values) == lengths


@pytest.mark.parametrize(
    "values,lengths",
    [
        ([], (0,)),
        ([1], (1,)),
        ([[]], (1, 0)),
        ([[1]], (1, 1)),
        ([[1, 2], [1]], (2, 2)),
        ([[[1], [2]], [[1]]], (2, 2, 1)),
        ([[[1, 2], [2]], [[1]]], (2, 2, 2)),
        ([[[1], [2]], [[1, 2]]], (2, 2, 2)),
        ([[[1], 2], [[1, 2]]], (2, 2)),
        ([[[1, 2]], [[1], 2]], (2, 2)),
    ],
)
def test_find_best_shape(values, lengths):
    assert ab.find_best_shape(values) == lengths


@pytest.mark.parametrize(
    "values,depth,flatted",
    [
        ([[1, 2], [3]], 2, [1, 2, 3]),
        ([[1, 2], []], 2, [1, 2]),
        ([[1, [2]], [3]], 2, [1, [2], 3]),
        ([[1, [2]], [3]], None, [1, 2, 3]),
        ([[[[1]]]], 1, [[[[1]]]]),  # ndim - 3
        ([[[[1]]]], 2, [[[1]]]),  # ndim - 2
        ([[[[1]]]], 3, [[1]]),  # ndim - 1
        ([[[[1]]]], 4, [1]),  # ndim
    ],
)
def test_flatten(values, depth, flatted):
    assert list(ab.flatten(values, depth)) == flatted


@pytest.mark.parametrize(
    "values,masked",
    [
        ([], []),
        ([1], [1]),
        ([[1, 2], [3]], [[1, 2], [3, 0]]),
        ([[1, 2], []], [[1, 2], [0, 0]]),
        ([[1, 2], [3], [4, 5, 6]], [[1, 2, 0], [3, 0, 0], [4, 5, 6]]),
        ([[[1, 2], [3]]], [[[1, 2], [3, 0]]]),
        ([[[1, 2], []]], [[[1, 2], [0, 0]]]),
        (
            [[[1, 2], [3]], [[1, 2, 3]]],
            [[[1, 2, 0], [3, 0, 0]], [[1, 2, 3], [0, 0, 0]]],
        ),
        (
            [[[[], [1, 2]]], [[[]], [[3]]]],
            [
                [[[0, 0], [1, 2]], [[0, 0], [0, 0]]],
                [[[0, 0], [0, 0]], [[3, 0], [0, 0]]],
            ],
        ),
    ],
)
def test_to_masked(values, masked):
    masked = np.array(masked)
    masked = np.ma.array(masked, mask=(masked == 0))
    array = ab.to_masked(values)
    np.testing.assert_array_equal(array, masked)
    np.testing.assert_array_equal(array.mask, masked.mask)


def test_to_masked_inconsistent_dims():
    with pytest.raises(ValueError):
        ab.to_masked([[1], 1])


@pytest.mark.parametrize(
    "values,mask",
    [
        ([], []),
        ([1], [1]),
        ([[1, 2], [3]], [[1, 1], [1, 0]]),
        ([[1, 2], []], [[1, 1], [0, 0]]),
        ([[1, 2], [3], [4, 5, 6]], [[1, 1, 0], [1, 0, 0], [1, 1, 1]]),
        ([[[1, 2], [3]]], [[[1, 1], [1, 0]]]),
        ([[[1, 2], []]], [[[1, 1], [0, 0]]]),
        (
            [[[1, 2], [3]], [[1, 2, 3]]],
            [[[1, 1, 0], [1, 0, 0]], [[1, 1, 1], [0, 0, 0]]],
        ),
        (
            [[[[], [1, 2]]], [[[]], [[3]]]],
            [
                [[[0, 0], [1, 2]], [[0, 0], [0, 0]]],
                [[[0, 0], [0, 0]], [[1, 0], [0, 0]]],
            ],
        ),
    ],
)
def test__mask(values, mask):
    mask = np.array(mask, dtype=bool)
    values = ab._mask(values, mask.shape)
    np.testing.assert_array_equal(values, mask)


@pytest.mark.xfail
def test__mask_shape_more_dims():
    shape = (1, 1, 1)
    mask = ab._mask([[1]], shape)
    assert mask.shape == shape


def test__mask_shape_less_dims():
    shape = (1, 1)
    mask = ab._mask([[[1]]], shape)
    assert mask.shape == shape


@pytest.mark.parametrize("shape", [(1, 2), (2, 1), (2, 2)])
def test__mask_shape_bigger(shape):
    mask = ab._mask([[1]], shape)
    assert mask.shape == shape


@pytest.mark.xfail
@pytest.mark.parametrize("shape", [(1, 2), (2, 1), (2, 2)])
def test__mask_shape_smaller(shape):
    mask = ab._mask([[1, 1], [1, 1]], shape)
    assert mask.shape == shape


def test_mask():
    values = mock.Mock()
    shape = mock.Mock()
    with mock.patch(
        "tesliper.glassware.array_base.find_best_shape", return_value=shape
    ), mock.patch("tesliper.glassware.array_base._mask"):
        ab.mask(values)
        ab.find_best_shape.assert_called_with(values)
        ab._mask.assert_called_with(values, shape)


@pytest.fixture
def class_array():
    class Cls:
        arr = ab.ArrayProperty(dtype=int)

    return Cls


@pytest.fixture
def class_array_check_x():
    class Cls:
        arr = ab.ArrayProperty(dtype=int, check_against="x")

    return Cls


@pytest.fixture
def class_array_xetters():
    class Cls:
        arr = ab.ArrayProperty(
            dtype=int, fget=mock.Mock(), fset=mock.Mock(), fdel=mock.Mock()
        )

    return Cls


def test_getter_called(class_array_xetters):
    arr = class_array_xetters()
    _ = arr.arr
    class_array_xetters.arr.fget.assert_called()
    class_array_xetters.arr.fget.assert_called_with(arr)


def test_setter_called(class_array_xetters):
    arr = class_array_xetters()
    arr.arr = 1
    class_array_xetters.arr.fset.assert_called()
    class_array_xetters.arr.fset.assert_called_with(arr, 1)


def test_deletter_called(class_array_xetters):
    arr = class_array_xetters()
    del arr.arr
    class_array_xetters.arr.fdel.assert_called()
    class_array_xetters.arr.fdel.assert_called_with(arr)


def test_sanitizer_called():
    class Cls:
        arr = ab.ArrayProperty(fsan=mock.Mock(return_value=[1, 2, 3]))

    arr = Cls()
    arr.arr = 1
    assert arr.arr.tolist() == [1, 2, 3]
    Cls.arr.fsan.assert_called()


def test_docs_from_getter():
    class Cls:
        def func(self):
            """Docstring for testing."""

        arr = ab.ArrayProperty(fget=func)

    assert Cls.arr.__doc__ == """Docstring for testing."""


def test_array_property_class_access(class_array):
    assert type(class_array.arr) is ab.ArrayProperty


def test_array_property_name(class_array):
    assert class_array.arr.name == "arr"


def test_array_property_check_input_no_attribute(class_array_check_x):
    with pytest.raises(AttributeError):
        class_array_check_x.arr.check_input(class_array_check_x(), [1, 2, 3])


def test_array_property_check_input_against(class_array_check_x):
    arr = class_array_check_x()
    arr.x = [1, 2, 3]
    assert class_array_check_x.arr.check_input(arr, [4, 5, 6]).tolist() == [4, 5, 6]


def test_array_property_check_input_against_deep(class_array_check_x):
    class_array_check_x.arr.check_depth = 2
    arr = class_array_check_x()
    arr.x = [[1], [2], [3]]
    assert class_array_check_x.arr.check_input(arr, [[4], [5], [6]]).tolist() == [
        [4],
        [5],
        [6],
    ]


def test_array_property_check_input_against_deeper(class_array_check_x):
    class_array_check_x.arr.check_depth = 2
    arr = class_array_check_x()
    arr.x = [[[1]], [[2]], [[3]]]
    assert class_array_check_x.arr.check_input(arr, [[4], [5], [6]]).tolist() == [
        [4],
        [5],
        [6],
    ]


def test_array_property_check_input_against_deep_unmatching(class_array_check_x):
    class_array_check_x.arr.check_depth = 2
    arr = class_array_check_x()
    arr.x = [[1], [2], [3]]
    with pytest.raises(ValueError):
        class_array_check_x.arr.check_input(arr, [1, 2, 3])


def test_array_property_check_input_against_deep_unmatching_allowed(
    class_array_check_x,
):
    class_array_check_x.arr.check_depth = 2
    arr = class_array_check_x()
    arr.allow_data_inconsistency = True
    arr.x = [[1], [2], [3]]
    assert class_array_check_x.arr.check_input(arr, [4, 5, 6]).tolist() == [4, 5, 6]


def test_array_property_check_input_unmatching(class_array_check_x):
    arr = class_array_check_x()
    arr.x = [1, 2]
    with pytest.raises(ValueError):
        class_array_check_x.arr.check_input(arr, [1, 2, 3])


def test_array_property_check_input_unmatching_allowed(class_array_check_x):
    arr = class_array_check_x()
    arr.x = [1, 2]
    arr.allow_data_inconsistency = True
    assert class_array_check_x.arr.check_input(arr, [1, 2, 3]).tolist() == [1, 2, 3]


def test_array_property_check_input_inconsistent(class_array_check_x):
    arr = class_array_check_x()
    arr.x = [1, 2]
    with pytest.raises(InconsistentDataError):
        class_array_check_x.arr.check_input(arr, [[1, 2], [3]])


def test_array_property_check_input_inconsistent_allowed(
    monkeypatch, class_array_check_x
):
    pad_mock = mock.Mock(return_value=[[1, 2], [3, 0]])
    monkeypatch.setattr(ab, "to_masked", pad_mock)
    arr = class_array_check_x()
    arr.x = [1, 2]
    arr.allow_data_inconsistency = True
    out = class_array_check_x.arr.check_input(arr, [[1, 2], [3]])
    pad_mock.assert_called_with([[1, 2], [3]], dtype=int, fill_value=0)
    assert out == [[1, 2], [3, 0]]


def test_array_property_decorator_with_dtype():
    mck = mock.Mock()

    class Cls:
        @ab.ArrayProperty(dtype=mck)
        def arr(self):
            pass

    assert Cls.arr.dtype is mck
    assert Cls.arr.fset is None
    assert Cls.arr.fdel is None
    assert Cls.arr.check_against is None


def test_array_property_decorator():
    class Cls:
        @ab.ArrayProperty
        def arr(self):
            pass

    assert Cls.arr.fset is None
    assert Cls.arr.fdel is None
    assert Cls.arr.check_against is None


single_value = {
    int: st.integers(min_value=-sys.maxsize, max_value=sys.maxsize),
    float: st.floats(allow_nan=False),
    str: st.characters(blacklist_categories=["Cs"], blacklist_characters=["\x00"]),
}
list_of_identical = {
    key: st.builds(lambda v, n: [v] * n, strat, st.integers(min_value=1, max_value=10))
    for key, strat in single_value.items()
}
list_of_unique = {
    key: st.lists(strat, min_size=2, unique=True) for key, strat in single_value.items()
}
list_of_identical_lists = {
    key: st.builds(lambda v, n: [v] * n, strat, st.integers(min_value=1, max_value=10))
    for key, strat in list_of_unique.items()
}
list_of_unique_lists = {
    key: st.builds(
        lambda vals: [list(v) for v in vals],
        st.integers(min_value=1, max_value=10).flatmap(
            lambda n: st.lists(
                st.tuples(*[strat for _ in range(n)]), min_size=2, unique=True
            ),
        ),
    )
    for key, strat in single_value.items()
}


@pytest.fixture
def mock_check_input(monkeypatch):
    monkeypatch.setattr(
        ab.ArrayProperty,
        "check_input",
        mock.Mock(side_effect=lambda i, x: ab.np.array(x)),
    )


@pytest.fixture(params=(int, float, str), scope="module")
def class_collapsable_array(request):
    class Cls:
        arr = ab.CollapsibleArrayProperty(dtype=request.param)

    return Cls


@pytest.fixture(params=(int, float, str), scope="module")
def class_strict_collapsable_array(request):
    class Cls:
        arr = ab.CollapsibleArrayProperty(dtype=request.param, strict=True)

    return Cls


@pytest.fixture(scope="module")
def instance_not_adi():
    return mock.Mock(allow_data_inconsistency=False)


@pytest.fixture(scope="module")
def instance_adi():
    return mock.Mock(allow_data_inconsistency=True)


@pytest.mark.usefixtures("mock_check_input")
@given(st.data())
def test_collapsable_single_value(class_collapsable_array, instance_not_adi, data):
    value = data.draw(single_value[class_collapsable_array.arr.dtype])
    assert class_collapsable_array.arr.check_input(
        instance_not_adi, value
    ).tolist() == [value]
    ab.ArrayProperty.check_input.assert_not_called()


@pytest.mark.usefixtures("mock_check_input")
@given(st.data(), st.integers(min_value=1, max_value=10))
def test_collapsable_list_of_identical(
    class_collapsable_array, instance_not_adi, data, number
):
    value = data.draw(single_value[class_collapsable_array.arr.dtype])
    arr = [value] * number
    assert class_collapsable_array.arr.check_input(instance_not_adi, arr).tolist() == [
        value
    ]
    ab.ArrayProperty.check_input.assert_called()


@pytest.mark.usefixtures("mock_check_input")
@given(st.data())
def test_strict_collapsable_list_of_unique(
    class_strict_collapsable_array, instance_not_adi, data
):
    values = data.draw(list_of_unique[class_strict_collapsable_array.arr.dtype])
    with pytest.raises(InconsistentDataError):
        class_strict_collapsable_array.arr.check_input(instance_not_adi, values)


@pytest.mark.usefixtures("mock_check_input")
@given(st.data())
def test_collapsable_list_of_unique_allowed(
    class_collapsable_array, instance_adi, data
):
    values = data.draw(list_of_unique[class_collapsable_array.arr.dtype])
    assert (
        class_collapsable_array.arr.check_input(instance_adi, values).tolist() == values
    )


@pytest.mark.usefixtures("mock_check_input")
def test_collapsable_empty_list(class_collapsable_array, instance_not_adi):
    assert class_collapsable_array.arr.check_input(instance_not_adi, []).tolist() == []


@pytest.mark.usefixtures("mock_check_input")
@given(st.data())
def test_cllapsable_list_of_identical_lists(
    class_collapsable_array, instance_not_adi, data
):
    values = data.draw(list_of_identical_lists[class_collapsable_array.arr.dtype])
    assert class_collapsable_array.arr.check_input(
        instance_not_adi, values
    ).tolist() == [values[0]]


@pytest.mark.usefixtures("mock_check_input")
@given(st.data())
def test_strict_cllapsable_list_of_unique_lists(
    class_strict_collapsable_array, instance_not_adi, data
):
    values = data.draw(list_of_unique_lists[class_strict_collapsable_array.arr.dtype])
    with pytest.raises(InconsistentDataError):
        class_strict_collapsable_array.arr.check_input(instance_not_adi, values)


@pytest.mark.usefixtures("mock_check_input")
@given(st.data())
def test_cllapsable_list_of_unique_lists_allowed(
    class_collapsable_array, instance_adi, data
):
    values = data.draw(list_of_unique_lists[class_collapsable_array.arr.dtype])
    assert (
        class_collapsable_array.arr.check_input(instance_adi, values).tolist() == values
    )


@pytest.fixture
def filenames_mock(monkeypatch):
    mocked = mock.PropertyMock()
    monkeypatch.setattr(ab.ArrayBase, "filenames", mocked)
    return mocked


@pytest.fixture
def values_mock(monkeypatch):
    mocked = mock.PropertyMock()
    monkeypatch.setattr(ab.ArrayBase, "values", mocked)
    return mocked


@pytest.fixture
def array_subclass():
    class Sub(ab.ArrayBase):
        associated_genres = ("bla",)

    return Sub


def test_array_base_subclass_no_associated_genres():
    with pytest.raises(TypeError):

        class Sub(ab.ArrayBase):
            pass


def test_array_base_subclass_str_associated_genres():
    with pytest.raises(TypeError):

        class Sub(ab.ArrayBase):
            associated_genres = "bla"


def test_array_base_get_args(filenames_mock, values_mock, array_subclass):
    values_mock.return_value = [1, 2, 3]
    filenames_mock.return_value = ["f1", "f2", "f3"]
    arr = array_subclass(genre="bla", filenames=["f1", "f2", "f3"], values=[1, 2, 3])
    assert arr.get_repr_args() == {
        "genre": "bla",
        "filenames": ["f1", "f2", "f3"],
        "values": [1, 2, 3],
        "allow_data_inconsistency": False,
    }


def test_array_base_get_args_not_stored_arg(filenames_mock, values_mock):
    class Sub(ab.ArrayBase):
        associated_genres = ()

        def __init__(
            self, genre, filenames, values, other, allow_data_inconsistency=False
        ):
            super().__init__(genre, filenames, values, allow_data_inconsistency)

    values_mock.return_value = [1, 2, 3]
    filenames_mock.return_value = ["f1", "f2", "f3"]
    arr = Sub(genre="bla", filenames=["f1", "f2", "f3"], values=[1, 2, 3], other="foo")
    assert arr.get_repr_args() == {
        "genre": "bla",
        "filenames": ["f1", "f2", "f3"],
        "values": [1, 2, 3],
        "other": None,
        "allow_data_inconsistency": False,
    }


def test_array_base_get_args_not_stored_arg_with_default(filenames_mock, values_mock):
    class Sub(ab.ArrayBase):
        associated_genres = ()

        def __init__(
            self,
            genre,
            filenames,
            values,
            other="foo",
            allow_data_inconsistency=False,
        ):
            super().__init__(genre, filenames, values, allow_data_inconsistency)

    values_mock.return_value = [1, 2, 3]
    filenames_mock.return_value = ["f1", "f2", "f3"]
    arr = Sub(genre="bla", filenames=["f1", "f2", "f3"], values=[1, 2, 3], other="foo")
    assert arr.get_repr_args() == {
        "genre": "bla",
        "filenames": ["f1", "f2", "f3"],
        "values": [1, 2, 3],
        "other": "foo",
        "allow_data_inconsistency": False,
    }


@pytest.fixture
def get_repr_args_mock(monkeypatch):
    monkeypatch.setattr(
        ab.ArrayBase,
        "get_repr_args",
        mock.Mock(
            return_value={
                "genre": "bla",
                "filenames": ["f1", "f2", "f3"],
                "values": [1, 2, 3],
                "allow_data_inconsistency": False,
            }
        ),
    )


def test_array_base_str(
    get_repr_args_mock, filenames_mock, values_mock, array_subclass
):
    values_mock.return_value = [1, 2, 3]
    m = mock.Mock()
    type(m).size = mock.PropertyMock(return_value=3)
    filenames_mock.return_value = m
    arr = array_subclass(genre="bla", filenames=["f1", "f2", "f3"], values=[1, 2, 3])
    assert str(arr) == "[Sub of genre 'bla', 3 conformers]"


def test_array_base_repr(
    get_repr_args_mock, filenames_mock, values_mock, array_subclass
):
    values_mock.return_value = [1, 2, 3]
    filenames_mock.return_value = ["f1", "f2", "f3"]
    arr = array_subclass(genre="bla", filenames=["f1", "f2", "f3"], values=[1, 2, 3])
    assert (
        repr(arr) == "Sub(genre='bla', filenames=['f1', 'f2', 'f3'], "
        "values=[1, 2, 3], allow_data_inconsistency=False)"
    )


def test_array_base_len(filenames_mock, values_mock, array_subclass):
    filenames_mock.return_value = ["f1", "f2", "f3"]
    arr = array_subclass(genre="bla", filenames=["f1", "f2", "f3"], values=[1, 2, 3])
    assert len(arr) == 3


def test_array_base_bool(filenames_mock, values_mock, array_subclass):
    type(filenames_mock).size = mock.PropertyMock(return_value=3)
    arr = array_subclass(genre="bla", filenames=["f1", "f2", "f3"], values=[1, 2, 3])
    assert bool(arr)


def test_array_base_bool_empty(filenames_mock, values_mock, array_subclass):
    m = mock.Mock()
    type(m).size = mock.PropertyMock(return_value=0)
    filenames_mock.return_value = m
    arr = array_subclass(genre="bla", filenames=[], values=[])
    assert arr.filenames.size == 0
    assert not bool(arr)



"""This module contains a definition of Abstract Base Class for file parsers."""
import re
from abc import ABC, abstractmethod
from typing import Callable, Dict, Iterable, Tuple, Type, Union

from ..exceptions import InvalidStateError

_PARSERS: Dict[str, Type["ParserBase"]] = {}


class ParserBase(ABC):
    """Abstract Base Class for parsers implemented as finite state machines.

    This base class defines some methods to organize work parsers implemented as finite
    state machines: automates registration of methods and functions as parser's states,
    manages its execution, and registers derived class as parser used for certain type
    of files (which registry is used by :class:`.Soxhlet` object).

    The default parsing flow goes as follow:

    1. method :meth:`.parse` is called with file handle as argument;
    2. method :meth:`.initial` is set as a 'workhorse'
    3. 'workhorse' is called for consecutive lines in file handle
    4. :meth:`.initial` checks if any registered trigger matches current line
    5. :meth:`.workhorse` is changed to method associated with first matching trigger
    6. calling 'workhorse' on consecutive lines continues
    7. :meth:`.parse` returns dictionary with extracted values

    To make this possible, each method marked as state should return dictionary
    (or sequence convertible to dict) and handle changing 'workhorse' to next
    appropriate state. To mark a method as parser's state use ParserBase.state
    decorator in class definition or add a state directly to parser instance
    using 'add_state' method.

    When subclassing ParserBase, one should implement :meth:`.initial` and
    :meth:`.parse` methods. Those abstract methods implement basic functionality,
    described above. See methods' documentation for more details. If you wish not to use
    default ParserBase's protocol, simply override those methods to your liking. Values
    for class attributes :attr:`.extensions` and :attr:`.purpose` should also be
    provided.

    To register class derived from ParserBase for use by :class:`.Soxhlet` object,
    simply set :attr:`.purpose` class attribute to name, under which class should be
    registered. Setting it to one of names already defined (e.g. 'gaussian') will
    override the default parser used by :class:`.Soxhlet` object.

    Attributes
    ----------
    states: dict
        Dictionary of parser states, created automatically on object
        instantiation from object methods marked as states; method name is used
        as a key by default.
    triggers: dict
        Dictionary of triggers for parser states, created automatically on
        object instantiation from object methods marked as states with triggers;
        key for a particular state trigger should be the same as state's key in
        :attr:`.states` dictionary.
    """

    @property
    @classmethod
    @abstractmethod
    def extensions(cls) -> Tuple[str]:
        return tuple()

    extensions.__doc__ = """File extensions that should be cosidered compatible with a
    parser subclassing :class:`.ParserBase`. It will be used by :class:`.Soxhlet` to
    identify which files to parse when reading files in batch. Should be a class
    attribute with a tuple of str, where each element is a file extension. May also be
    an empty tuple, if files discovery feature is not needed for the parser.
    """

    @property
    @classmethod
    @abstractmethod
    def purpose(cls) -> str:
        return ""

    purpose.__doc__ = """An identifier for a parser subclassing :class:`.ParserBase`. It
    allows ``tesliper`` to pick a correct parser for each parsing task. A falsy value,
    i.e. an empty string or ``None`` prevents the parser from beeing registered for use
    by ``tesliper``. If custom subclass uses a *purpose* already known, e.g. "gaussian"
    or "spectra", it will override the original parser for this purpose.
    """

    def __init__(self):
        self.states = {}
        self.triggers = {}
        states = (
            (name, method)
            for name, method in (
                (n, getattr(self, n)) for n in dir(self) if n != "workhorse"
            )
            if hasattr(method, "is_state") and method.is_state
        )
        for name, method in states:
            self.add_state(method, name, getattr(method, "trigger", None))
        self.workhorse = self.initial

    def __init_subclass__(cls, **kwargs):
        global _PARSERS
        if cls.purpose is ParserBase.purpose:
            raise TypeError(
                f"`{cls.__name__}` must provide `purpose` class attribute. "
                "It may be an empty string if this class should not be registered."
            )
        if cls.extensions is ParserBase.extensions:
            raise TypeError(
                f"`{cls.__name__}` must provide `extensions` class attribute. "
                "It may be an empty tuple if no extensions should be associated with "
                "this class."
            )
        if cls.purpose:
            _PARSERS[cls.purpose] = cls
        if not hasattr(cls.initial, "is_state"):
            cls.initial.is_state = True

    @property
    def workhorse(self) -> Callable:
        """Callable marked as a current state used by parser object.

        Setter can take a callable or a string as a parameter. If name as
        string is passed to setter, it will be translated to a method
        registered as state. If no method was registered under this name,
        :class:`.InvalidStateError` will be raised. No other checks are performed
        when argument is callable."""
        return self._workhorse

    @workhorse.setter
    def workhorse(self, state: Union[Callable, str]):
        if callable(state):
            self._workhorse = state
        else:
            try:
                self._workhorse = self.states[state]
            except KeyError:
                raise InvalidStateError(
                    f"{state} is not callable nor registered state name"
                )

    def add_state(self, state: Callable, name: str = "", trigger: str = ""):
        """Register callable as parser's state.

        This method registers a callable under *name* key in :attr:`.states`
        dictionary. If *trigger* parameter is given, it is registered under the
        same key in :attr:`.triggers` dictionary.

        Parameters
        ----------
        state: Callable
            callable, that is to be registered as parser's state
        name: str, optional
            name under which the callable should be registered;
            defaults to callable.__name__
        trigger: str, optional
            string with regular expression, that will be compiled with re module

        Returns
        -------
        Callable
            callable object registered as state
        """
        if not name:
            name = state.__name__
        self.states[name] = state
        if trigger:
            self.triggers[name] = re.compile(trigger)
        elif hasattr(state, "trigger"):
            self.triggers[name] = re.compile(state.trigger)
        return state

    def remove_state(self, name: str):
        """Removes the state from parser's registered states.

        Parameters
        ----------
        name : str
            name of state, that should be unregistered

        Raises
        ------
        InvalidStateError
            if no callable was registered under the name 'name'
        """
        if name not in self.states:
            raise InvalidStateError(f"No state registered under name: {name}.")
        del self.states[name]
        if name in self.triggers:
            del self.triggers[name]

    @abstractmethod
    def initial(self, line: str) -> dict:
        """An initial parser state.

        A default implementation checks if any of defined triggers matches
        a line and sets an associated state as parser's workhorse, if it does.
        This is an abstract method and should be overridden in subclass.
        Its default implementation can be used, however, by calling
        ``super().initial(line)`` in subclass's method.

        Notes
        -----
        :meth:`.initial` method is always registered as parser's state.

        Parameters
        ----------
        line: str
            currently parsed line

        Returns
        -------
        dict
            empty dictionary"""
        for name, reg in self.triggers.items():
            match = reg.match(line)
            if match:
                self.workhorse = self.states[name]
                return {}
        return {}

    @abstractmethod
    def parse(self, lines: Iterable) -> dict:
        """Parses consecutive elements of iterable and returns data found
        as dictionary.

        Dictionary with extracted data is updated with workhorse's return value,
        so all states should return dictionary or compatible sequence.
        This is an abstract method and should be overridden in subclass.
        Its default implementation can be used, however, by calling
        ``data = super().parse(lines)`` in subclass's method.

        Notes
        -----
        After execution - either successful or interrupted by exception -
        :attr:`.workhorse` is set back to :meth:`.initial` method.

        Parameters
        ----------
        lines: Iterable
            iterable (i.e. file handle), that will be parsed, line by line

        Returns
        -------
        dict
            dictionary with data extracted by parser

        Raises
        ------
        InvalidStateError
            if dictionary can't be updated with state's return value"""
        data = {}
        try:
            for line in lines:
                output = self.workhorse(line)
                try:
                    data.update(output)
                except TypeError as error:
                    raise InvalidStateError(
                        f"State {self.workhorse} should return value "
                        f'convertible to a dictionary, not "{type(output)}".'
                    ) from error
                except ValueError as error:
                    raise InvalidStateError(
                        f"Value returned by state {self.workhorse} could not "
                        f"be converted to dictionary."
                    ) from error
        except Exception:
            raise
        finally:
            self.workhorse = self.initial
        return data

    @staticmethod
    def state(state=None, trigger=None):
        """Convenience decorator for registering a method as parser's state.
        It can be with or without 'trigger' parameter, like this:

        >>> @ParserBase.state
        ... def method(self, arg): pass

        or

        >>> @ParserBase.state(trigger='triggering regex')
        ... def method(self, arg): pass

        This function marks a method *state* as parser's state by defining ``is_state``
        attribute on said method and setting its values to ``True``. If *trigger* is
        given, it is stored in method's attribute *trigger*. During instantiation of
        :class:`.ParserBase`'s subclass, methods marked as states are registered under
        ``method.__name__`` key in its :attr:`.states` (and possibly :attr:`.triggers`)
        attribute. It is meaningless if used outside of :class:`.ParserBase`'s subclass
        definition.

        Parameters
        ----------
        state: Callable
            callable, that is to be registered as parser's state
        trigger: str, optional
            string with regular expression, that will be compiled with re module

        Returns
        -------
        Callable
            callable object registered as state if 'state' was given or
            decorator if only 'trigger' was given

        Raises
        ------
        TypeError
            if no arguments given
        InvalidStateError
            if *state* argument is not callable
        """
        if callable(state):
            state.is_state = True
            if trigger:
                state.trigger = trigger
            return state
        elif state is None and trigger is None:
            raise TypeError("At least one argument must be given.")
        elif trigger is None:
            raise InvalidStateError(
                f"'state' argument should be callable, not {type(state)}"
            )
        else:
            return lambda s, t=trigger: ParserBase.state(s, t)



"""Converters between string and integer representations of atoms."""

from enum import IntEnum
from typing import List, Union

import numpy as np

from ..exceptions import InvalidElementError

_atoms = (
    "H He Li Be B C N O F Ne Na Mg Al Si P S Cl Ar K Ca Sc Ti V Cr Mn Fe Co Ni Cu "
    "Zn Ga Ge As Se Br Kr Rb Sr Y Zr Nb Mo Tc Ru Rh Pd Ag Cd In Sn Sb Te I Xe Cs Ba "
    "La Ce Pr Nd Pm Sm Eu Gd Tb Dy Ho Er Tm Yb Lu Hf Ta W Re Os Ir Pt Au Hg Tl Pb "
    "Bi Po At Rn Fr Ra Ac Th Pa U Np Pu Am Cm Bk Cf Es Fm Md No Lr Rf Db Sg Bh Hs Mt "
    "Ds Rg Cn Nh Fl Mc Lv Ts Og"
)
atomicnums = {at: num + 1 for num, at in enumerate(_atoms.split())}
atoms_symbols = {v: k for k, v in atomicnums.items()}
Atom = IntEnum("Atom", _atoms)
Atom.__doc__ = """An enumeration that maps symbols of atoms to respective atomic numbers.

This enumeration is introduced for your convenience: whenever you need to reference
an atom by its atomic number, you may use appropriate symbol-value of this Enum instead.
Providing e.g. ``Atom.Au`` rather than an integer 79 for Au's atomic number is probably
a bit easier and definitely more readable.
"""


def symbol_of_element(element: Union[int, str]) -> str:
    """Returns symbol of given element. If *element* is a symbol of an element
    already, it is capitalized and returned (so input's letters case doesn't
    matter).

    Parameters
    ----------
    element: int or str
        element's atomic number

    Returns
    -------
    str
        symbol of an element

    Raises
    ------
    ValueError
        when *element* is not a whole number or cannot be converted to integer
    TypeError
        if *element* cannot be interpreted as integer
    InvalidElementError
        if *element* is not an atomic number of any known element"""
    stringified = str(element).capitalize()
    if stringified in atomicnums:
        return stringified
    try:
        integerized = int(element)
    except (ValueError, TypeError) as exc:
        if isinstance(element, str):
            raise ValueError(f"Cannot convert element {element} to integer.") from exc
        raise TypeError(
            f'Type "{type(element)}" cannot be interpreted as integer.'
        ) from exc
    if isinstance(element, float) and not element == integerized:
        raise ValueError(
            f"Element's atomic number should be a whole number, " f"{element} given."
        )
    elif integerized in atoms_symbols:
        return atoms_symbols[integerized]
    else:
        raise InvalidElementError(f"Unknown element: {element}")


def atomic_number(element: Union[int, str]) -> int:
    """Returns atomic number of given element. If *element* is an atomic number
    already, it is returned without change.

    Parameters
    ----------
    element: str or int
        element's symbol or atomic number (letters case doesn't matter if
        string is given)

    Returns
    -------
    int
        atomic number of an element

    Raises
    ------
    InvalidElementError
        when *element* cannot be converted to element's atomic number
    TypeError
        if *element* cannot be interpreted as integer or string"""
    stringified = str(element).capitalize()
    if stringified in atomicnums:
        return atomicnums[stringified]
    elif element in atoms_symbols:
        return element
    elif isinstance(element, (str, int, float, np.str_, np.integer, np.float64)):
        raise InvalidElementError(f"Unknown element: {element}")
    else:
        raise TypeError(f"Expected str or int, got '{type(element)}'.")


def validate_atoms(atoms: Union[int, str, List[Union[int, str]]]) -> List[int]:
    """Checks if given *atoms* represent a list of valid atom identifiers
    (symbols or atomic numbers). Returns list of atomic numbers of those atoms
    if it does or rises an exception if it doesn't.

    Parameters
    ----------
    atoms: int, str or iterable of int or str
        Atoms to validate. Atoms as space-separated string are also accepted.

    Returns
    -------
    list of int
        List of given atoms' atomic numbers.

    Raises
    ------
    InvalidElementError
        if *atoms* cannot be interpreted as list of atoms' identifiers"""
    if isinstance(atoms, (str, np.str_)):
        atoms = atoms.split()
    elif isinstance(atoms, (int, np.integer, float, np.float64)):
        atoms = [atoms]
    try:
        return [atomic_number(a) for a in atoms]
    except (InvalidElementError, TypeError) as exc:
        raise InvalidElementError(
            f"Cannot interpret {atoms} as list of atoms' identifiers."
        ) from exc



from unittest import mock

import numpy as np
from hypothesis import given, strategies as st

from tesliper.datawork import energies as en
import pytest


def test_calculate_deltas_wrong_type_string():
    with pytest.raises(ValueError):
        en.calculate_deltas("string")


def test_calculate_deltas_wrong_type_integer():
    with pytest.raises(TypeError):
        en.calculate_deltas(1)


def test_calculate_deltas_not_iterable():
    with pytest.raises(TypeError):
        en.calculate_deltas(object())


def test_calculate_deltas_empty():
    out = en.calculate_deltas([])
    assert out.tolist() == []


def test_calculate_deltas_one_value():
    out = en.calculate_deltas([1])
    assert out.tolist() == [0]


def test_calculate_deltas_convertible_string():
    out = en.calculate_deltas(["1"])
    assert out.tolist() == [0]


@given(st.lists(st.floats(allow_nan=False, allow_infinity=False), min_size=1))
def test_calculate_deltas_more_values(data):
    m = min(data)
    out = en.calculate_deltas(data)
    assert out.tolist() == [d - m for d in data]


@pytest.fixture
def deltas(monkeypatch):
    monkeypatch.setattr(en, "calculate_deltas", mock.Mock())


def test_calculate_min_factors_empty(deltas):
    en.calculate_deltas.return_value = np.array([])
    out = en.calculate_min_factors([])
    assert out.tolist() == []


def test_calculate_min_factors_one_value(deltas):
    en.calculate_deltas.return_value = np.array([0])
    out = en.calculate_min_factors([1])
    assert out.tolist() == [1]


def test_calculate_min_factors_more_values(deltas):
    en.calculate_deltas.return_value = np.array([0, 0.6275095, 2 * 0.6275095])
    out = en.calculate_min_factors([0, 0.001, 0.002])
    assert np.round(out, decimals=8).tolist() == [1, 0.34676265, 0.12024433]


def test_calculate_min_factors_other_temp(deltas):
    en.calculate_deltas.return_value = np.array([0, 0.6275095, 2 * 0.6275095])
    out = en.calculate_min_factors([0, 0.001, 0.002], t=200)
    assert np.round(out, decimals=8).tolist() == [1, 0.20620689, 0.04252128]


@pytest.fixture
def factors(monkeypatch):
    monkeypatch.setattr(en, "calculate_min_factors", mock.Mock())


def test_calculate_populations_empty(factors):
    en.calculate_min_factors.return_value = np.array([])
    out = en.calculate_populations([])
    assert out.tolist() == []


def test_calculate_populations_one_value(factors):
    en.calculate_min_factors.return_value = np.array([1])
    out = en.calculate_populations([1])
    assert out.tolist() == [1]


def test_calculate_populations_more_values(factors):
    en.calculate_min_factors.return_value = np.array([1, 0.34676265, 0.12024433])
    out = en.calculate_populations([0, 0.001, 0.002])
    assert np.round(out, decimals=8).tolist() == [0.68166002, 0.23637423, 0.08196575]


def test_calculate_populations_other_temp(factors):
    en.calculate_min_factors.return_value = np.array([1, 0.20620689, 0.04252128])
    out = en.calculate_populations([0, 0.001, 0.002], t=200)
    assert np.round(out, decimals=8).tolist() == [0.8008148, 0.16513353, 0.03405167]



"""Optical activity to signal intensity converters."""

import logging as lgg

import numpy as np

# LOGGER
logger = lgg.getLogger(__name__)
logger.setLevel(lgg.DEBUG)


# GLOBAL VARIABLES
DEFAULT_ACTIVITIES = {
    "ir": "dip",
    "vcd": "rot",
    "uv": "vosc",
    "ecd": "vrot",
    "raman": "raman1",
    "roa": "roa1",
}


# MODULE FUNCTIONS
def dip_to_ir(values: np.ndarray, frequencies: np.ndarray) -> np.ndarray:
    """Calculates signal intensity of IR spectrum.

    Parameters
    ----------
    values: numpy.ndarray
        Dipole strength values extracted from gaussian output files.
    frequencies: numpy.ndarray
        Frequencies extracted from gaussian output files.

    Returns
    -------
    numpy.ndarray
        List of calculated intensity values."""
    return values * frequencies * 0.010886


def rot_to_vcd(values: np.ndarray, frequencies: np.ndarray) -> np.ndarray:
    """Calculates signal intensity of VCD spectrum.

    Parameters
    ----------
    values: numpy.ndarray
        Rotator strength values extracted from gaussian output files.
    frequencies: numpy.ndarray
        Frequencies extracted from gaussian output files.

    Returns
    -------
    numpy.ndarray
        List of calculated intensity values."""
    return values * frequencies * 0.0435441


def osc_to_uv(values: np.ndarray) -> np.ndarray:
    """Calculates signal intensity of UV spectrum.

    Parameters
    ----------
    values: numpy.ndarray
        Oscillator strength values extracted from gaussian output files.

    Returns
    -------
    numpy.ndarray
        List of calculated intensity values."""
    return values * 2.315351857e08


def rot_to_ecd(values: np.ndarray, wavelengths: np.ndarray) -> np.ndarray:
    """Calculates signal intensity of ECD spectrum.

    Parameters
    ----------
    values: numpy.ndarray
        Rotator strength values extracted from gaussian output files.
    wavelengths: numpy.ndarray
        Wavelengths extracted from gaussian output files.

    Returns
    -------
    numpy.ndarray
        List of calculated intensity values."""
    return values * wavelengths * 0.0435441


def dip_to_uv(values: np.ndarray, wavelengths: np.ndarray) -> np.ndarray:
    """Calculates signal intensity of UV spectrum.

    Parameters
    ----------
    values: numpy.ndarray
        Dipole strength values extracted from gaussian output files.
    wavelengths: numpy.ndarray
        Wavelengths extracted from gaussian output files.

    Returns
    -------
    numpy.ndarray
        List of calculated intensity values."""
    return values * wavelengths * 0.010886



import pytest
from hypothesis import given, strategies as st
from tesliper.datawork import atoms


# test symbol_of_element
def test_symbol_of_elem_atomic_number_int():
    assert atoms.symbol_of_element(1) == "H"


def test_symbol_of_elem_atomic_number_float():
    assert atoms.symbol_of_element(1.0) == "H"


def test_symbol_of_elem_atomic_number_str():
    assert atoms.symbol_of_element("1") == "H"


def test_symbol_of_elem_not_atomic_number_int():
    with pytest.raises(atoms.InvalidElementError):
        atoms.symbol_of_element(0)


@pytest.mark.parametrize("element", [[], {}, tuple(), set()])
def test_symbol_of_elem_cannot_convert(element):
    with pytest.raises(TypeError):
        atoms.symbol_of_element(element)


def test_symbol_of_elem_not_atomic_number_float():
    with pytest.raises(ValueError):
        atoms.symbol_of_element(0.5)


def test_symbol_of_elem_not_atomic_number_str():
    with pytest.raises(atoms.InvalidElementError):
        atoms.symbol_of_element("0")


def test_symbol_of_elem_element_symbol():
    assert atoms.symbol_of_element("H") == "H"


def test_symbol_of_elem_element_symbol_lowercase():
    assert atoms.symbol_of_element("he") == "He"


def test_symbol_of_elem_not_element_symbol():
    with pytest.raises(ValueError):
        atoms.symbol_of_element("bla")


# test atomic_number
def test_atomic_number_valid_symbol():
    assert atoms.atomic_number("H") == 1


def test_atomic_number_lowercase_symbol():
    assert atoms.atomic_number("he") == 2


def test_atomic_number_invalid_symbol():
    with pytest.raises(atoms.InvalidElementError):
        atoms.atomic_number("bla")


def test_atomic_number_atomic_number():
    assert atoms.atomic_number(1) == 1


def test_atomic_number_not_atomic_number():
    with pytest.raises(atoms.InvalidElementError):
        atoms.atomic_number(0)


@pytest.mark.parametrize("element", [[], {}, tuple(), set()])
def test_atomic_number_cannot_convert(element):
    with pytest.raises(TypeError):
        atoms.atomic_number(element)


# test validate_atoms
def test_validate_integer():
    assert atoms.validate_atoms(1) == [1]


def test_validate_float():
    assert atoms.validate_atoms(2.0) == [2]


def test_validate_string():
    assert atoms.validate_atoms("B") == [5]


def test_validate_string_with_spaces():
    assert atoms.validate_atoms("H H C") == [1, 1, 6]


def test_validate_list_of_integers():
    assert atoms.validate_atoms([1, 1, 6]) == [1, 1, 6]


def test_validate_list_of_floats():
    assert atoms.validate_atoms([1.0, 6.0]) == [1, 6]


def test_validate_list_of_strings():
    assert atoms.validate_atoms(["H", "H"]) == [1, 1]


@pytest.mark.parametrize("element", [[], {}, tuple(), set()])
def test_validate_cannot_convert(element):
    with pytest.raises(atoms.InvalidElementError):
        atoms.validate_atoms([element])



from unittest.mock import Mock

import numpy as np
import pytest
from hypothesis import assume, given
from hypothesis import strategies as st

from tesliper.datawork import spectra as sp


# test count_imaginary
def test_count_imaginary_empty_array():
    assert sp.count_imaginary(np.array([])).tolist() == []


def test_count_imaginary_single_value_positive():
    assert sp.count_imaginary(np.array([1])).tolist() == 0


def test_count_imaginary_single_value_zero():
    assert sp.count_imaginary(np.array([0])).tolist() == 0


def test_count_imaginary_single_value_negative():
    assert sp.count_imaginary(np.array([-1])).tolist() == 1


def test_count_imaginary_zero_dimensions_positive():
    assert sp.count_imaginary(np.array(1, dtype=int)).tolist() == 0


def test_count_imaginary_zero_dimensions_zero():
    assert sp.count_imaginary(np.array(0, dtype=int)).tolist() == 0


def test_count_imaginary_zero_dimensions_negative():
    assert sp.count_imaginary(np.array(-1, dtype=int)).tolist() == 1


def test_count_imaginary_zero_imag_one_conf():
    assert sp.count_imaginary(np.arange(0, 6, 1).reshape(1, -1)).tolist() == [0]


def test_count_imaginary_one_imag_one_conf():
    assert sp.count_imaginary(np.arange(-1, 6, 1).reshape(1, -1)).tolist() == [1]


def test_count_imaginary_zero_imag_two_confs():
    assert sp.count_imaginary(np.arange(0, 6, 1).reshape(2, -1)).tolist() == [0, 0]


def test_count_imaginary_one_imag_two_confs():
    assert sp.count_imaginary(np.arange(-1, 5, 1).reshape(2, -1)).tolist() == [1, 0]


def test_count_imaginary_two_imag_two_confs():
    assert sp.count_imaginary(np.array([[1, -1, 1], [-1, 1, 1]])).tolist() == [1, 1]


def test_count_imaginary_tree_dimensions():
    with pytest.raises(ValueError):
        sp.count_imaginary(np.array([[[1]]]))


# test find_imaginary
def test_find_imaginary_less_dim_zero_imag():
    assert sp.find_imaginary(np.arange(0, 6, 1).reshape(1, -1)).tolist() == []


def test_find_imaginary_less_dim_one_imag():
    assert sp.find_imaginary(np.arange(-1, 5, 1).reshape(1, -1)).tolist() == [0]


def test_find_imaginary_zero_imag():
    assert sp.find_imaginary(np.arange(0, 6, 1).reshape(2, -1)).tolist() == []


def test_find_imaginary_imag_at_first():
    assert sp.find_imaginary(np.arange(-1, 5, 1).reshape(2, -1)).tolist() == [0]


def test_find_imaginary_imag_at_second():
    assert sp.find_imaginary(np.array([[1, 1, 1], [1, -1, 1]])).tolist() == [1]


def test_find_imaginary_more_dimensions():
    with pytest.raises(ValueError):
        sp.find_imaginary(np.array([[[1]]]))


# test gaussian
def test_gaussian_mean_0_std_dev_1():
    np.testing.assert_array_almost_equal(
        sp.gaussian(np.array([1]), np.array([0]), np.arange(-5, 6, 1), np.sqrt(2)),
        [
            0.0000014867195147343,
            0.0001338302257648850,
            0.0044318484119380100,
            0.0539909665131881000,
            0.2419707245191430000,
            0.3989422804014330000,
            0.2419707245191430000,
            0.0539909665131881000,
            0.0044318484119380100,
            0.0001338302257648850,
            0.0000014867195147343,
        ],  # values from Excel 2010 NORM.DIST
    )


def test_gaussian_width_zero():
    with pytest.raises(ValueError):
        sp.gaussian(np.array([1]), np.array([0]), np.arange(-5, 6, 1), 0)


def test_gaussian_width_below_zero():
    with pytest.raises(ValueError):
        sp.gaussian(np.array([1]), np.array([0]), np.arange(-5, 6, 1), -1)


def test_gaussian_unequal_sizes():
    with pytest.raises(ValueError):
        sp.gaussian(np.array([1, 2]), np.array([0]), np.arange(-5, 6, 1), 1)


def test_gaussian_empty_intensities():
    out = sp.gaussian(np.array([]), np.array([]), np.arange(-5, 6, 1), 1)
    assert out.tolist() == np.zeros(11).tolist()


def test_gaussian_empty_abscissia():
    out = sp.gaussian(np.array([1]), np.array([0]), np.array([]), 1)
    assert out.tolist() == []


# test lorentzian
def test_lorentzian_mean_0_std_dev_1():
    np.testing.assert_array_almost_equal(
        sp.lorentzian(np.array([1]), np.array([0]), np.arange(-5, 6, 1), 1 / 2),  # L/2
        [
            0.00630316606304536,
            0.00979415034411664,
            0.01720593979371840,
            0.03744822190397540,
            0.12732395447351600,
            0.63661977236758100,
            0.12732395447351600,
            0.03744822190397540,
            0.01720593979371840,
            0.00979415034411664,
            0.00630316606304536,
        ]  # values from Excel 2010
        # =(L/(2*PI())) / (POWER((x-x0); 2) + POWER((L/2); 2))
    )


def test_lorentzian_width_zero():
    with pytest.raises(ValueError):
        sp.lorentzian(np.array([1]), np.array([0]), np.arange(-5, 6, 1), 0)


def test_lorentzian_width_below_zero():
    with pytest.raises(ValueError):
        sp.lorentzian(np.array([1]), np.array([0]), np.arange(-5, 6, 1), -1)


def test_lorentzian_unequal_sizes():
    with pytest.raises(ValueError):
        sp.lorentzian(np.array([1, 2]), np.array([0]), np.arange(-5, 6, 1), 1)


def test_lorentzian_empty_intensities():
    out = sp.lorentzian(np.array([]), np.array([]), np.arange(-5, 6, 1), 1)
    assert out.tolist() == np.zeros(11).tolist()


def test_lorentzian_empty_abscissia():
    out = sp.lorentzian(np.array([1]), np.array([0]), np.array([]), 1)
    assert out.tolist() == []


# test calculate_spectra


@pytest.fixture
def fitting():
    mock = Mock()

    def func(intensities, frequencies, abscissa, width, m=mock):
        return np.ones_like(abscissa, dtype=int) * m.call_count

    mock.side_effect = func
    return mock


def test_calculate_spectra_zero_conformers(fitting):
    out = sp.calculate_spectra(np.array([]), np.array([]), np.arange(3), 1, fitting)
    assert out.tolist() == []


def test_calculate_spectra_one_conformer(fitting):
    out = sp.calculate_spectra(
        np.ones((1, 2)), np.ones((1, 2)), np.arange(3), 1, fitting
    )
    assert out.tolist() == [[1, 1, 1]]


def test_calculate_spectra_two_conformers(fitting):
    out = sp.calculate_spectra(
        np.ones((2, 2)), np.ones((2, 2)), np.arange(3), 1, fitting
    )
    assert out.tolist() == [[1, 1, 1], [2, 2, 2]]


def test_calculate_spectra_unmatching_arrays(fitting):
    with pytest.raises(ValueError):
        sp.calculate_spectra(np.ones((2, 2)), np.ones((3, 2)), np.arange(3), 1, fitting)


def test_calculate_spectra_empty_abscissa(fitting):
    out = sp.calculate_spectra(
        np.ones((2, 2)), np.ones((2, 2)), np.array([]), 1, fitting
    )
    assert out.tolist() == [[], []]


# test calculate_average
def test_calculate_average_one_dim():
    out = sp.calculate_average([2, 6], [0.25, 0.75])
    assert out == 5


def test_calculate_average_two_dims():
    out = sp.calculate_average([[2, 4], [6, 8]], [0.25, 0.75])
    assert out.tolist() == [5, 7]


def test_calculate_average_three_dims():
    out = sp.calculate_average([[[2, 4], [6, 8]], [[6, 8], [6, 12]]], [0.25, 0.75])
    assert out.tolist() == [[5, 7], [6, 11]]


def test_calculate_average_normalize():
    out = sp.calculate_average([2, 6], [1, 3])
    assert out == 5


def test_calculate_average_unmatching_sizes():
    with pytest.raises(ValueError):
        sp.calculate_average([1, 2], [1, 2, 3])
    with pytest.raises(ValueError):
        sp.calculate_average([1, 2, 3], [1, 2])
    with pytest.raises(ValueError):
        sp.calculate_average([1, 2, 3], 1)


@given(st.integers(min_value=3, max_value=100), st.integers(), st.integers())
def test_offset_one_peak_same_size(size, peak, shift):
    # TODO: when size == 2 and shift == 1 this gives offset == 1 instead of -1
    #       investigate on this error
    assume(0 <= peak < size)
    assume(0 <= peak + shift < size)
    a, b = np.zeros(size), np.zeros(size)
    a[peak] = 10
    b[peak + shift] = 10
    offset = sp.idx_offset(a, b)
    assert offset == -shift


@given(
    st.floats(min_value=0.1, max_value=1),
    st.floats(min_value=0.1, max_value=1),
    st.integers(min_value=100, max_value=1000),
    st.integers(min_value=100, max_value=1000),
    st.booleans(),
    st.booleans(),
)
def test_unify_abscissa(d1, d2, n1, n2, up, decreasing_abscissa):
    x1 = np.arange(n1) * d1
    x2 = np.arange(n2) * d2
    delta = min([d1, d2]) if up else max([d1, d2])
    if decreasing_abscissa:
        x1, x2 = x1[::-1], x2[::-1]
        d1, d2, delta = -d1, -d2, -delta
    ax, ay, bx, by = sp.unify_abscissa(x1, np.arange(n1), x2, np.arange(n2), up)
    assert np.allclose([ax[1] - ax[0], bx[1] - bx[0]], [delta])


@given(
    st.floats(min_value=0.1, max_value=10),
    st.integers(min_value=100, max_value=1000),
    st.integers(min_value=0, max_value=999),
    st.integers(min_value=0, max_value=999),
)
def test_find_offset(delta, size, peak1, peak2):
    assume(peak1 < size and peak2 < size)
    ay, by = np.zeros(size), np.zeros(size)
    ay[peak1], by[peak2] = 1, 1
    x = np.arange(0, delta * size, delta)
    offset = sp.find_offset(x, ay, x, by)
    assert offset == delta * (peak1 - peak2)


@given(
    st.lists(st.floats(allow_nan=False, allow_infinity=False)),
    st.lists(st.floats(allow_nan=False, allow_infinity=False)),
)
def test_find_scaling(a, b):
    scale = sp.find_scaling(a, b)
    assert scale >= 0



"""Functions for calculating populations of conformers based on their relative energies
using Boltzmann distribution.
"""

# IMPORTS
import logging as lgg

import numpy as np

# LOGGER
logger = lgg.getLogger(__name__)
logger.setLevel(lgg.DEBUG)


# GLOBAL VARIABLES
BOLTZMANN = 0.0019872041
"""Value of Boltzmann constant in kcal/(mol*K)."""

HARTREE_TO_KCAL_PER_MOL = 627.5095
"""Multiply by this factor to convert from Hartree/mol to kcal/mol."""


# MODULE FUNCTIONS
def calculate_deltas(energies):
    """Calculates energy difference between each conformer and lowest energy conformer.

    Parameters
    ----------
    energies : numpy.ndarray or iterable of float
        List of conformers energies.

    Returns
    -------
    numpy.ndarray
        List of energy differences from lowest energy."""
    array = np.asanyarray(energies, dtype=float)
    if not array.shape:
        raise TypeError(f"Expected iterable, not '{type(energies)}'")
    # need to reshape if single value given instead of sequence
    try:
        return array - array.min()
    except ValueError:
        # if no values, return empty array.
        return np.array([])


def calculate_min_factors(energies, t=298.15):
    """Calculates list of conformers' Boltzmann factors respective to lowest
    energy conformer in system of given temperature.

    Notes
    -----
    Boltzmann factor of two states is defined as:

    .. math::

        \frac{F(state_1)}{F(state_2)} = e^{(E_2 - E_1)/kt}

    where :math:`E_1` and :math:`E_2` are energies of states 1 and 2,
    :math:`k` is Boltzmann constant, :math:`k = 0.0019872041 kcal/(mol*K)`,
    and :math:`t` is temperature of the system.

    Parameters
    ----------
    energies : numpy.ndarray or iterable
        List of conformers energies in kcal/mol units.
    t : float, optional
        Temperature of the system in K, defaults to 298,15 K.

    Returns
    -------
    numpy.ndarary
        List of conformers' Boltzmann factors respective to lowest
        energy conformer."""
    arr = -calculate_deltas(energies)
    return np.exp(arr / (t * BOLTZMANN))


def calculate_populations(energies, t=298.15):
    """Calculates Boltzmann distribution of conformers of given energies.

    Parameters
    ----------
    energies : numpy.ndarray or iterable
        List of conformers energies in kcal/mol units.
    t : float, optional
        Temperature of the system in K, defaults to 298,15 K.

    Returns
    -------
    numpy.ndarary
        List of conformers populations calculated as Boltzmann distribution."""
    arr = calculate_min_factors(energies, t)
    return arr / arr.sum()



"""A tool for batch parsing files from specified directory."""
import logging as lgg
from pathlib import Path
from typing import Any, Generator, Iterable, List, Optional, Set, Tuple, Union

from .parser_base import _PARSERS

# LOGGER
logger = lgg.getLogger(__name__)
logger.setLevel(lgg.DEBUG)


# CLASSES
# TODO: Consider integration with gauopen interface: http://gaussian.com/interfacing/
# TODO: supplement class docstring
class Soxhlet:
    """A tool for data extraction from files in specific directory. Typical use:

    >>> s = Soxhlet('absolute/path_to/working/directory')
    >>> data = s.extract()
    """

    def __init__(
        self,
        path: Optional[Union[str, Path]] = None,
        purpose: str = "gaussian",
        wanted_files: Optional[Iterable[Union[str, Path]]] = None,
        extension: Optional[str] = None,
        recursive: bool = False,
    ):
        """
        Parameters
        ----------
        path : str or pathlib.Path
            String representing absolute path to directory containing files,
            which will be the subject of data extraction.
        purpose : str
            Determines which from registered parsers should be used for extraction.
            *purpose*\\s supported out-of-the-box are "gaussian", "spectra", and
            "parameters".
        wanted_files : list of str or pathlib.Path objects, optional
            List of files, that should be loaded for further extraction. If
            omitted, all output files present in directory will be processed.
        extension : str, optional
            A string representing file extension of output files, that should be
            parsed. If omitted, Soxhlet will try to resolve it based on
            contents of directory given in *path* parameter.
        recursive : bool
            If True, given *path* will be searched recursively, extracting data from
            subdirectories, otherwise subdirectories are ignored and only files
            placed directly in *path* will be parsed.

        Raises
        ------
        FileNotFoundError
            If path passed as argument to constructor doesn't exist
            or is not a directory.
        ValueError
            If no parser is registered for given *purpose*.
        """
        self.path = path
        self.wanted_files = wanted_files
        self.extension = extension
        self.recursive = recursive
        try:
            self.parser = _PARSERS[purpose]()
        except KeyError:
            raise ValueError(f"Unknown purpose: {purpose}.")

    @property
    def path(self) -> Path:
        return self._path

    @path.setter
    def path(self, value: Union[str, Path]):
        value = Path() if value is None else Path(value)
        if not value.is_dir():
            raise FileNotFoundError(f"Path not found: {value}")
        self._path = value.resolve()

    @property
    def all_files(self):
        """List of all files present in directory bounded to Soxhlet instance.
        If its *recursive* attribute is ``True``, also files from subdirectories
        are included."""
        iterable = self.path.iterdir() if not self.recursive else self.path.rglob("*")
        return [v for v in iterable if v.is_file()]

    @property
    def files(self):
        """List of all wanted files available in given directory. If wanted_files
        is not specified, evaluates to all files in said directory. If Soxhlet
        object's *recursive* attribute is ``True``, also files from subdirectories
        are included."""
        wanted_empty = not self.wanted_files
        return [
            f for f in self.all_files if wanted_empty or f.stem in self.wanted_files
        ]

    @property
    def wanted_files(self) -> Optional[Set[str]]:
        """Set of files that are desired for data extraction, stored as filenames
        without an extension. Any iterable of strings or Path objects is transformed
        to this form.

        >>> s = Soxhlet()
        >>> s.wanted_files = [Path("./dir/file_one.out"), Path("./dir/file_two.out")]
        >>> s.wanted_files
        {"file_one", "file_two"}

        May also be set to ``None`` or other "falsy" value, in such case it is ignored.
        """
        return self._wanted_files

    @wanted_files.setter
    def wanted_files(self, files: Optional[Iterable[Union[str, Path]]]):
        self._wanted_files = None if not files else {Path(f).stem for f in files}

    @property
    def output_files(self) -> List[Path]:
        """List of (sorted by file name) gaussian output files from files
        list associated with Soxhlet instance.
        """
        try:
            ext = self.extension if self.extension else self.guess_extension()
        except (ValueError, FileNotFoundError) as error:
            logger.warning(f"{error} Returning empty list.")
            return []
        try:
            gf = sorted(self.filter_files(ext))
        except ValueError:
            gf = []
        return gf

    def filter_files(self, ext: Optional[str] = None) -> List[Path]:
        """Filters files from filenames list.

        Filters file names in list associated with :class:`.Soxhlet` object
        instance. It returns list of file names ending with provided ext
        string, representing file extension and starting with any of filenames
        associated with instance as wanted_files if those were provided.

        Parameters
        ----------
        ext : str
            Strings representing file extension.

        Returns
        -------
        list
            List of filtered filenames as strings.

        Raises
        ------
        ValueError
            If parameter *ext* is not given and attribute :attr:`.extension` in
            ``None``.
        """
        ext = ext if ext is not None else self.extension
        if ext is None:
            raise ValueError(
                "Parameter *ext* must be given if attribute *extension* is *None*."
            )
        filtered = [f for f in self.files if f.name.endswith(ext)]
        return filtered

    def guess_extension(self) -> str:
        """Tries to figure out which extension should be assumed.

        Looks for files, which names end with one of the extensions defined by currently
        used parser. Returns extension that matches as the only one. Raises an exception
        if extension cannot be easily guessed.

        Returns
        -------
        str
            The extension of files that are present in filenames list, which current
            parser can parse.

        Raises
        ------
        ValueError
            If more than one type of files declared by a current parser as possibly
            compatible is present in list of filenames.
        FileNotFoundError
            If none of files declared by a current parser as possibly compatible are
            present in list of filenames.
        TypeError
            If current parser does not declare any compatible file extensions.
        """
        if not self.parser.extensions:
            raise TypeError(
                f"Cannot guess extension: parser {type(self.parser).__name__} "
                "does not define any compatible extensions."
            )
        available = tuple(
            ext
            for ext in self.parser.extensions
            if any(f.name.endswith(ext) for f in self.all_files)
        )
        if len(available) > 1:
            raise ValueError(
                f"{', '.join(f'.{a}' for a in available)} files mixed in directory."
            )
        elif not available:
            raise FileNotFoundError(
                "Didn't found any of "
                f"{', '.join(f'.{e}' for e in self.parser.extensions)} files."
            )
        else:
            return available[0]

    def extract_iter(self) -> Generator[Tuple[str, dict], None, None]:
        """Extracts data from files associated with :class:`.Soxhlet` instance (*via*
        :attr:`.path` and :attr:`.wanted_files` attributes), using a current parser
        (determined by a *purpose* provided on :class:`.Soxhlet`'s instantiation).
        Implemented as generator. If Soxhlet instance's :attr:`.recursive` attribute is
        ``True``, also files from subdirectories are parsed.

        Yields
        ------
        tuple
            Two item tuple with name of parsed file as first and extracted
            data as second item, for each file associated with Soxhlet instance.
        """
        for file in self.output_files:
            logger.debug(f"Starting extraction from file: {file}")
            with file.open() as handle:
                data = self.parser.parse(handle)
            logger.debug("file done.\n")
            yield file.stem, data

    def extract(self) -> dict:
        """Extracts data from files associated with :class:`.Soxhlet` instance (*via*
        :attr:`.path` and :attr:`.wanted_files` attributes), using a current parser
        (determined by a *purpose* provided on :class:`.Soxhlet`'s instantiation). If
        :attr:`.Soxhlet.recursive` attribute is ``True``, also files from subdirectories
        are parsed.

        Returns
        ------
        dict of dicts
            dictionary of extracted data, with name of parsed file as key and
            data as value, for each file associated with Soxhlet instance.
        """
        return {f: d for f, d in self.extract_iter()}

    def parse_one(self, source: Union[str, Path]) -> Any:
        """Parse one file using current parser (determined by a *purpose* provided on
        :class:`.Soxhlet`'s instantiation) and return extracted data.

        Parameters
        ----------
        source : str or Path
            Path or Path-like object to a file. May be given as an absolute path or
            relative to the :attr:`.Soxhlet.path`.

        Returns
        -------
        any
            Data in a format that current parser provides.

        Raises
        ------
        FileNotFoundError
            If no *source* file is found.
        """
        appended = self.path / source
        if appended.is_file():
            path = appended
        elif Path(source).is_file():
            path = Path(source)
        else:
            raise FileNotFoundError(f"Cannot find such file: '{source}'.")
        return self.parser.parse(path)



import logging
import sys
import tkinter as tk
import tkinter.ttk as ttk

# LOGGER
logger = logging.getLogger(__name__)


def is_scrollable(widget):
    scrollable = False
    for scrollcommand in ["xscrollcommand", "yscrollcommand"]:
        try:
            _ = widget[scrollcommand]
        except tk.TclError:
            logger.debug(f"Widget {widget} does not offer '{scrollcommand}' config.")
        except TypeError:
            logger.debug(f"Widget {widget} not subscriptable with config keys.")
        else:
            scrollable = True
    try:
        disabled = str(widget["state"]) == tk.DISABLED
    except tk.TclError:
        # state irrelevant
        logger.debug(f"Widget {widget} does not offer 'state' config.")
        disabled = False
    return scrollable and not disabled


# CLASSES
class AutoScrollbar(ttk.Scrollbar):
    """A scrollbar that hides automatically when entire canvas is visible.
    Must be placed with `.grid()`.
    """

    def set(self, low, high):
        if float(low) <= 0.0 and float(high) >= 1.0:
            self.grid_remove()
        else:
            self.grid()  # restores widget previously removed from grid
        super().set(low, high)

    def pack(self, *_args, **_kwargs):
        raise tk.TclError("cannot use pack with this widget")

    def place(self, *_args, **_kwargs):
        raise tk.TclError("cannot use place with this widget")


class ScrollableFrame(ttk.Frame):
    """A frame, that allows you to scroll content vertically.
    Shows scrollbar only when it is necessary."""

    def __init__(self, parent, content_cls=ttk.Frame):
        super().__init__(parent)
        self.canvas = tk.Canvas(self, highlightthickness=0)
        self.scrollbar = AutoScrollbar(self, command=self.canvas.yview)
        self.canvas.configure(yscrollcommand=self.scrollbar.set)
        self.content = content_cls(self.canvas)
        self.content.bind("<Configure>", self._on_content_configure)
        self.canvas.create_window((0, 0), window=self.content, anchor="nw")
        self.canvas.grid(row=0, column=0, sticky="nwse")
        self.scrollbar.grid(row=0, column=1, sticky="nws")
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.canvas.columnconfigure(0, weight=1)
        self.canvas.rowconfigure(0, weight=1)

        self._scheduled = None
        self._hold_scrolling = False
        self._hold_delay = 500

        # Scroll with mouse wheel when cursor over canvas
        self.bind("<Enter>", self._bound_to_mousewheel)
        self.bind("<Leave>", self._unbound_to_mousewheel)

        # Enable delayed scrolling hold on other scrollable widgets
        self.bind_all("<Enter>", self._on_child_enter, "+")
        self.bind_all("<Leave>", self._on_child_leave, "+")

    def _on_child_enter(self, event):
        """If entered widget is ScrollableFrame's child and scrollable,
        schedule passing scroll control to this widget."""
        command = self._child_reverse_bindtags(event.widget)
        if not command:
            self._scheduled = self.after(self._hold_delay, self._on_child_stop, event)

    def _on_child_leave(self, event):
        """If widget left before scheduled hold was executed, cancel it and revert
        changes. Give scroll control to ScrollableFrame."""
        if self._scheduled is not None:
            self._child_reverse_bindtags(event.widget)
            self.after_cancel(self._scheduled)
            self._scheduled = None
        self._hold_scrolling = False

    def _on_child_stop(self, event):
        """Give scroll control to widget over the mouse cursor."""
        self._child_reverse_bindtags(event.widget)
        self._scheduled = None
        self._hold_scrolling = True

    def _child_reverse_bindtags(self, widget):
        """Reverse order of widget's bindtags to control if widget-specific or "all"
        bindings are executed first. Do it only if widget is a child of ScrollableFrame.
        """
        error = None
        if not str(widget).startswith(str(self.canvas)):
            error = f"widget is not a child of {self}"
        elif not isinstance(widget, tk.Widget):
            error = f"{widget} is not a tkinter.Widget object"
        elif not is_scrollable(widget):
            error = "non-scrollable or disabled widget"
        if error is not None:
            logger.debug(f"Bindtags reverse aborted: {error}.")
            return "ok"  # or maybe "continue"?

        reversed_ = widget.bindtags()[::-1]
        widget.bindtags(reversed_)
        logger.debug(f"New bindtags order: {reversed_}.")

    def _bound_to_mousewheel(self, event):
        self.canvas.bind_all("<MouseWheel>", self._on_mousewheel)
        # For Linux
        self.canvas.bind_all("<Button-4>", self._on_mousewheel)
        self.canvas.bind_all("<Button-5>", self._on_mousewheel)

    def _unbound_to_mousewheel(self, event):
        self.canvas.unbind_all("<MouseWheel>")
        self.canvas.unbind_all("<Button-4>")
        self.canvas.unbind_all("<Button-5>")

    def _on_mousewheel(self, event):
        logger.debug(f"Event caught by {self}._on_mousewheel handler.")
        # prevent scrolling when content fully visible
        # or when scrolling is held by other scrollable widget
        if not self._hold_scrolling and not self.scrollbar.get() == (0.0, 1.0):
            delta = (
                event.delta
                if sys.platform == "darwin"
                else int(-1 * (event.delta / 120))
            )
            self.canvas.yview_scroll(delta, "units")
            # don't propagate event to other bindtags
            # to enable delayed hold on scrollable widgets
            return "break"

    def _on_content_configure(self, event):
        self.canvas.configure(scrollregion=self.canvas.bbox("all"))
        # keep width of canvas as small as possible
        self.canvas.configure(width=event.width)



"""Data export to excel files."""
import logging as lgg
from itertools import chain, repeat, zip_longest
from pathlib import Path
from string import Template
from typing import Iterable, List, Optional, Sequence, Union

import numpy as np
import openpyxl as oxl

from ..glassware.arrays import (
    Bands,
    DataArray,
    Energies,
    FloatArray,
    InfoArray,
    SpectralActivities,
    SpectralData,
    Transitions,
)
from ..glassware.spectra import SingleSpectrum, Spectra
from .writer_base import WriterBase, _GenericArray

# LOGGER
logger = lgg.getLogger(__name__)
logger.setLevel(lgg.DEBUG)


# CLASSES
class XlsxWriter(WriterBase):
    """Writes extracted data to .xlsx file."""

    extension = "xlsx"

    def __init__(
        self,
        destination: Union[str, Path],
        mode: str = "x",
        filename: str = "tesliper-output.${ext}",
    ):
        """
        Parameters
        ----------
        destination : str or pathlib.Path
            Directory, to which generated files should be written.
        mode : str
            Specifies how writing to file should be handled. Should be one of
            characters: 'a' (append to existing file), 'x' (only write if file doesn't
            exist yet), or 'w' (overwrite file if it already exists).
        filename : str or string.Template
            Filename of created .xlsx file or a template for generation of the name
            using :meth:`.make_name` method.
        """
        super().__init__(destination=destination, mode=mode)
        file = self.destination / Template(filename).substitute(ext=self.extension)
        self.file = self.check_file(file)
        self.manual_save = False
        if self.mode == "a":
            self.workbook = oxl.load_workbook(self.file)
        else:
            self.workbook = oxl.Workbook()
            self.workbook.remove(self.workbook.active)

    def write(self, data: List) -> None:
        self.manual_save, orig_manual = True, self.manual_save
        try:
            super().write(data)
            if not orig_manual:
                self.workbook.save(self.file)
        finally:
            self.manual_save = orig_manual

    def generic(
        self,
        data: List[_GenericArray],
        name_template: Union[str, Template] = "${cat}.${det}",
    ):
        """Writes generic data from multiple :class:`.DataArray`-like objects to a
        single sheet. Said objects should provide a single value for each conformer.

        Parameters
        ----------
        data
            :class:`.DataArray` objects that are to be exported.
        name_template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.
        """
        wb = self.workbook
        genres = [arr.genre for arr in data]
        headers = ["Gaussian output file"] + [self._header[genre] for genre in genres]
        values = [arr.values for arr in data]
        lines = list(zip(data[0].filenames, *values))
        types = [type(arr).__name__.lower().replace("array", "") for arr in data]
        detail = "various" if len(set(types)) > 1 else types[0]
        genre = "misc" if len(genres) > 1 else genres[0]
        template_params = {
            "cat": "generic",
            "conf": "multiple",
            "det": detail,
            "genre": genre,
        }
        ws = wb.create_sheet(title=self.make_name(name_template, **template_params))
        ws.append(headers)
        for line in lines:
            ws.append(line)
        if not self.manual_save:
            wb.save(self.file)
        logger.info("Generic export to xlsx files done.")

    def overview(
        self,
        energies: Sequence[Energies],
        frequencies: Optional[DataArray] = None,
        stoichiometry: Optional[InfoArray] = None,
        name_template: Union[str, Template] = "${cat}",
    ):
        """Writes summarized information from multiple :class:`.Energies` objects to
        xlsx file. Creates a worksheet with energy values and calculated populations for
        each energy object given, as well as number of imaginary frequencies and
        stoichiometry of conformers if *frequencies* and *stoichiometry* are provided,
        respectively.

        Parameters
        ----------
        energies: list of glassware.Energies
            Energies objects that are to be exported
        frequencies: glassware.DataArray, optional
            DataArray object containing frequencies
        stoichiometry: glassware.InfoArray, optional
            InfoArray object containing stoichiometry information
        name_template : str or string.Template
            Template that will be used to generate filenames, defaults to "${cat}".
            Refer to :meth:`.make_name` documentation for details on supported
            placeholders.
        """
        wb = self.workbook
        template_params = {"cat": "overview", "conf": "multiple"}
        ws = wb.create_sheet(title=self.make_name(name_template, **template_params))
        ens_no = len(energies)
        headers = ["Gaussian output file", "Populations / %", "Energies / hartree"]
        headers += ["Imag"] if frequencies is not None else []
        headers += ["Stoichiometry"] if stoichiometry is not None else []
        cells = [
            "A1",
            "B1",
            f"{chr(66+ens_no)}1",
            f"{chr(66+2*ens_no)}1",
            f"{chr(67+2*ens_no)}1",
        ]
        for header, cell in zip(headers, cells):
            ws[cell] = header
        names = [self._header[en.genre] for en in energies]
        ws.append([""] + names + names)
        ws.merge_cells("A1:A2")
        ws.merge_cells(f"B1:{chr(65+ens_no)}1")
        ws.merge_cells(f"{chr(66+ens_no)}1:{chr(65+2*ens_no)}1")
        if frequencies is not None or stoichiometry is not None:
            ws.merge_cells(f"{chr(66+2*ens_no)}1:{chr(66+2*ens_no)}2")
        if frequencies is not None and stoichiometry is not None:
            ws.merge_cells(f"{chr(67+2*ens_no)}1:{chr(67+2*ens_no)}2")
        ws.freeze_panes = "A3"
        filenames = energies[0].filenames
        fmts = (
            ["0"]
            + ["0.00%"] * len(energies)
            + ["0." + "0" * (8 if en.genre == "scf" else 6) for en in energies]
            + ["0", "0"]
        )
        values = [en.values for en in energies]
        populs = [en.populations for en in energies]
        imag = frequencies.imaginary if frequencies is not None else []
        stoich = stoichiometry.values if stoichiometry is not None else []
        rows = zip_longest(filenames, *populs, *values, imag, stoich)
        for row_num, values in enumerate(rows):
            filtered_values = ((f, v) for f, v in zip(fmts, values) if v is not None)
            for col_num, (fmt, value) in enumerate(filtered_values):
                cell = ws.cell(row=row_num + 3, column=col_num + 1)
                cell.value = value
                cell.number_format = fmt
        # set cells width
        widths = [0] + [10] * ens_no + [16] * ens_no
        widths += [6] if frequencies is not None else []
        widths += [0] if stoichiometry is not None else []
        for column, width in zip(ws.columns, widths):
            if not width:
                width = max(len(str(cell.value)) for cell in column) + 2
            column_letter = oxl.utils.get_column_letter(column[0].column)
            ws.column_dimensions[column_letter].width = width
        if not self.manual_save:
            wb.save(self.file)
        logger.info("Overview export to xlsx files done.")

    def energies(
        self,
        energies: Energies,
        corrections: Optional[FloatArray] = None,
        name_template: Union[str, Template] = "distribution-${genre}",
    ):
        """Writes detailed information from multiple :class:`.Energies` objects to xlsx
        file. Creates one worksheet for each :class:`.Energies` object provided. The
        sheet contains energy values, energy difference to lowest-energy conformer,
        Boltzmann factor, population of each conformer and corrections, if those are
        provided.

        Parameters
        ----------
        energies: list of glassware.Energies
            Energies objects that are to be exported
        corrections: list of glassware.DataArray
            DataArray objects containing energies corrections
        name_template : str or string.Template
            Template that will be used to generate filenames, defaults to
            "distribution-${genre}". Refer to :meth:`.make_name` documentation for
            details on supported placeholders.
        """
        wb = self.workbook
        fmts = (
            ["0", "0.00%"]
            + ["0.0000"] * 2
            + ["0.00000000" if energies.genre == "scf" else "0.000000"] * 2
        )
        template_params = {
            "conf": "multiple",
            "genre": energies.genre,
            "cat": "populations",
        }
        ws = wb.create_sheet(title=self.make_name(name_template, **template_params))
        ws.freeze_panes = "A2"
        header = [
            "Gaussian output file",
            "Population / %",
            "Min. B. Factor",
            "DE / (kcal/mol)",
            "Energy / Hartree",
        ]
        header += ["Correction / Hartree"] if corrections is not None else []
        ws.append(header)
        corr_values = corrections.values if corrections is not None else []
        rows = zip_longest(
            energies.filenames,
            energies.populations,
            energies.min_factors,
            energies.deltas,
            energies.values,
            corr_values,
        )
        for row_num, values in enumerate(rows):
            filtered_values = ((f, v) for f, v in zip(fmts, values) if v is not None)
            for col_num, (fmt, value) in enumerate(filtered_values):
                cell = ws.cell(row=row_num + 2, column=col_num + 1)
                cell.value = value
                cell.number_format = fmt
        # set cells width
        widths = [0, 15, 14, 15, 16, 19]
        for column, width in zip(ws.columns, widths):
            if not width:
                width = max(len(str(cell.value)) for cell in column) + 2
            column_letter = oxl.utils.get_column_letter(column[0].column)
            ws.column_dimensions[column_letter].width = width
        if not self.manual_save:
            wb.save(self.file)
        logger.info("Energies export to xlsx files done.")

    def spectral_data(
        self,
        band: SpectralActivities,
        data: Iterable[SpectralData],
        name_template: Union[str, Template] = "${conf}.${cat}-${det}",
    ):
        """Writes :class:`.SpectralData` objects to xlsx file (one sheet for each
        conformer).

        Parameters
        ----------
        band: glassware.SpectralActivities
            object containing information about band at which transitions occur;
            it should be frequencies for vibrational data and wavelengths or
            excitation energies for electronic data
        data: iterable of glassware.SpectralData
            SpectralData objects that are to be serialized; all should contain
            information for the same conformers. Assumes that all *data*'s elements have
            the same *spectra_type*, which is passed to the *name_template* as "det".
        name_template : str or string.Template
            Template that will be used to generate filenames, defaults to
            "${conf}.${cat}-${det}".  Refer to :meth:`.make_name` documentation for
            details on supported placeholders.

        Raises
        ------
        ValueError
            if *data* is an empty sequence
        """
        self._spectral(
            band=band, data=data, name_template=name_template, category="data"
        )

    def spectral_activities(
        self,
        band: SpectralActivities,
        data: Iterable[SpectralActivities],
        name_template: Union[str, Template] = "${conf}.${cat}-${det}",
    ):
        """Writes :class:`.SpectralActivities` objects to xlsx file (one sheet for each
        conformer).

        Parameters
        ----------
        band: glassware.SpectralActivities
            object containing information about band at which transitions occur;
            it should be frequencies for vibrational data and wavelengths or
            excitation energies for electronic data
        data: iterable of glassware.SpectralActivities
            SpectralActivities objects that are to be serialized; all should contain
            information for the same conformers. Assumes that all *data*'s elements have
            the same *spectra_type*, which is passed to the *name_template* as "det".
        name_template : str or string.Template
            Template that will be used to generate filenames, defaults to
            "${conf}.${cat}-${det}". Refer to :meth:`.make_name` documentation for
            details on supported placeholders.

        Raises
        ------
        ValueError
            if *data* is an empty sequence
        """
        self._spectral(
            band=band, data=data, name_template=name_template, category="activities"
        )

    def _spectral(
        self,
        band: SpectralActivities,
        data: Union[Iterable[SpectralData], Iterable[SpectralActivities]],
        name_template: Union[str, Template],
        category: str,
    ):
        """Writes spectral data to .xlsx file (one sheet for each conformer).

        Parameters
        ----------
        band: glassware.SpectralActivities
            object containing information about band at which transitions occur;
            it should be frequencies for vibrational data and wavelengths or
            excitation energies for electronic data
        data: iterable of glassware.SpectralActivities or iterable of SpectralData
            SpectralActivities or SpectralData objects that are to be serialized; all
            should contain information for the same conformers. Assumes that all
            *data*'s elements have the same *spectra_type*, which is passed to the
            *name_template* as "det".
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.
        category : str
            Category of exported data genres.

        Raises
        ------
        ValueError
            if *data* is an empty sequence
        """
        wb = self.workbook
        data = [band] + list(data)
        try:
            spectra_type = data[1].spectra_type
        except IndexError:
            raise ValueError("No data to export.")
        genres = [bar.genre for bar in data]
        headers = [self._header[genre] for genre in genres]
        widths = [max(len(h), 10) for h in headers]
        fmts = [self._excel_formats[genre] for genre in genres]
        values = list(zip(*[bar.values for bar in data]))
        template_params = {"genre": band.genre, "cat": category, "det": spectra_type}
        for num, (fname, values_) in enumerate(zip(data[0].filenames, values)):
            template_params.update({"conf": fname, "num": num})
            ws = wb.create_sheet(title=self.make_name(name_template, **template_params))
            ws.append(headers)
            ws.freeze_panes = "B2"
            for column, width in zip(ws.columns, widths):
                column_letter = oxl.utils.get_column_letter(column[0].column)
                ws.column_dimensions[column_letter].width = width
            for col_num, (vals, fmt) in enumerate(zip(values_, fmts)):
                for row_num, v in enumerate(vals):
                    cell = ws.cell(row=row_num + 2, column=col_num + 1)
                    cell.value = v
                    cell.number_format = fmt
        if not self.manual_save:
            wb.save(self.file)
        logger.info("SpectralActivities export to xlsx files done.")

    def spectra(
        self,
        spectra: Spectra,
        name_template: Union[str, Template] = "${genre}",
    ):
        """Writes given spectra collectively to one sheet of xlsx workbook.

        Parameters
        ----------
        spectra: glassware.Spectra
            Spectra object, that is to be serialized
        name_template : str or string.Template
            Template that will be used to generate filenames, defaults to "${genre}".
            Refer to :meth:`.make_name` documentation for details on supported
            placeholders.
        """
        wb = self.workbook
        template_params = {"genre": spectra.genre, "cat": "spectra"}
        ws = wb.create_sheet(title=self.make_name(name_template, **template_params))
        ws.freeze_panes = "B2"
        A0 = spectra.units["x"]
        ws.append([A0] + list(spectra.filenames))
        title = (
            f"{spectra.genre} calculated with peak width = "
            f'{spectra.width} {spectra.units["width"]} and '
            f"{spectra.fitting} fitting, shown as "
            f'{spectra.units["x"]} vs. {spectra.units["y"]}'
        )
        ws["A1"].comment = oxl.comments.Comment(title, "Tesliper")
        for line in zip(spectra.x, *spectra.y):
            ws.append(line)
        if not self.manual_save:
            wb.save(self.file)
        logger.info("Spectra export to xlsx file done.")

    def single_spectrum(
        self,
        spectrum: SingleSpectrum,
        name_template: Union[str, Template] = "${cat}.${genre}-${det}",
    ):
        """Writes SingleSpectrum object to new sheet of xlsx workbook.

        Parameters
        ----------
        spectrum: glassware.SingleSpectrum
            spectrum, that is to be serialized
        name_template : str or string.Template
            Template that will be used to generate sheet names, defaults to
            "${cat}.${genre}-${det}". Refer to :meth:`.make_name` documentation for
            details on supported placeholders.
        """
        # TODO: add comment as in txt export
        wb = self.workbook
        template_params = {
            "genre": spectrum.genre,
            "cat": "spectrum",
            "det": spectrum.averaged_by,
        }
        ws = wb.create_sheet(title=self.make_name(name_template, **template_params))
        ws.append([spectrum.units["x"], spectrum.units["y"]])
        for row in zip(spectrum.x, spectrum.y):
            ws.append(row)
        if not self.manual_save:
            wb.save(self.file)
        logger.info("Spectrum export to xlsx files done.")

    def transitions(
        self,
        transitions: Transitions,
        wavelengths: Bands,
        only_highest=True,
        name_template: Union[str, Template] = "${conf}.${cat}-${det}",
    ):
        """Writes electronic transitions data to xlsx file
        (one sheet for each conformer).

        Parameters
        ----------
        transitions : glassware.Transitions
            Electronic transitions data that should be serialized.
        wavelengths : glassware.ElectronicActivities
            Object containing information about wavelength at which transitions occur.
        only_highest : bool
            Specifies if only transition of highest contribution to given band should
            be reported. If ``False`` all transition are saved to file.
            Defaults to ``True``.
        name_template : str or string.Template
            Template that will be used to generate filenames, defaults to
            "${conf}.${cat}-${det}". Refer to :meth:`.make_name` documentation for
            details on supported placeholders.
        """
        transtions_data = (
            transitions.highest_contribution
            if only_highest
            else (
                transitions.ground,
                transitions.excited,
                transitions.values,
                transitions.contribution,
            )
        )
        wb = self.workbook
        headers = [
            self._header[wavelengths.genre],
            "Ground",
            "Excited",
            "Coefficient",
            "Contribution",
        ]
        widths = [len(h) for h in headers]
        fmts = [self._excel_formats[wavelengths.genre], "0", "0", "0.0000", "0%"]

        template_params = {
            "genre": transitions.genre,
            "cat": "transitions",
            "det": "highest" if only_highest else "all",
        }
        for num, (fname, grounds, exciteds, values, contribs, bands) in enumerate(
            zip(
                transitions.filenames,
                *transtions_data,
                wavelengths.wavelen,
            )
        ):
            template_params.update({"conf": fname, "num": num})
            ws = wb.create_sheet(title=self.make_name(name_template, **template_params))
            ws.append(headers)
            ws.freeze_panes = "B2"
            for column, width in zip(ws.columns, widths):
                column_letter = oxl.utils.get_column_letter(column[0].column)
                ws.column_dimensions[column_letter].width = width
            row_num = 1
            for g, e, v, c, b in zip(grounds, exciteds, values, contribs, bands):
                try:
                    values_ = [
                        # print wavelength value only once
                        d
                        for d in zip(chain([b], repeat(None)), g, e, v, c)
                        # omit entry if any value is masked
                        if all(x is not np.ma.masked for x in d)
                    ]
                except TypeError:
                    # transition_data is transitions.highest_contribution
                    values_ = [(b, g, e, v, c)]
                for vals in values_:
                    row_num += 1
                    for col_num, (v_, fmt) in enumerate(zip(vals, fmts), start=1):
                        cell = ws.cell(row=row_num, column=col_num)
                        cell.value = v_
                        cell.number_format = fmt
        if not self.manual_save:
            wb.save(self.file)
        logger.info("Transitions export to xlsx files done.")



"""Data export to text files."""
import logging as lgg
from itertools import zip_longest
from string import Template
from typing import List, Optional, Sequence, Union

import numpy as np

from ..glassware.arrays import (
    Bands,
    Energies,
    FloatArray,
    InfoArray,
    SpectralActivities,
    SpectralData,
    Transitions,
)
from ..glassware.spectra import SingleSpectrum, Spectra
from .writer_base import WriterBase, _GenericArray

# LOGGER
logger = lgg.getLogger(__name__)
logger.setLevel(lgg.DEBUG)


# CLASSES
class TxtWriter(WriterBase):
    """Writes extracted or calculated data to .txt format files."""

    extension = "txt"

    def generic(
        self,
        data: List[_GenericArray],
        name_template: Union[str, Template] = "${cat}.${det}.${ext}",
    ):
        """Writes generic data from multiple :class:`.DataArray`-like objects to a
        single file. Said objects should provide a single value for each conformer.

        Parameters
        ----------
        data
            :class:`.DataArray` objects that are to be exported.
        name_template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.
        """
        genres = [arr.genre for arr in data]
        headers = ["Gaussian output file"] + [self._header[genre] for genre in genres]
        formatters = [
            self._formatters[g] if g in self._formatters else "{}" for g in genres
        ]
        values = [arr.values for arr in data]
        formatted = [[f.format(v) for v in vs] for f, vs in zip(formatters, values)]
        lines = list(zip(data[0].filenames, *formatted))
        widths = [max([len(v) for v in vs]) for vs in zip(headers, *lines)]
        types = [type(arr).__name__.lower().replace("array", "") for arr in data]
        detail = "various" if len(set(types)) > 1 else types[0]
        genre = "misc" if len(genres) > 1 else genres[0]
        template_params = {
            "cat": "generic",
            "conf": "multiple",
            "det": detail,
            "genre": genre,
        }
        headers_line = "   ".join([f"{e:<{w}}" for e, w in zip(headers, widths)])
        side = ["<"] + [">"] * len(genres)
        with self._get_handle(name_template, template_params) as handle:
            handle.write(headers_line + "\n")
            handle.write("-" * len(headers_line) + "\n")
            for line in lines:
                handle.write(
                    "   ".join([f"{e:{s}{w}}" for e, s, w in zip(line, side, widths)])
                )
                handle.write("\n")

    def overview(
        self,
        energies: Sequence[Energies],
        frequencies: Optional[Bands] = None,
        stoichiometry: Optional[InfoArray] = None,
        name_template: Union[str, Template] = "${cat}.${ext}",
    ):
        """Writes essential information from multiple Energies objects to
        single txt file.

        Notes
        -----
        All Energy objects given should contain information for the same set of files.

        Parameters
        ----------
        energies: list of glassware.Energies
            Energies objects that is to be exported
        frequencies: glassware.DataArray, optional
            DataArray object containing frequencies, needed for imaginary
            frequencies count
        stoichiometry: glassware.InfoArray, optional
            InfoArray object containing stoichiometry information
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.
        """
        filenames = energies[0].filenames
        imaginary = [] if frequencies is None else frequencies.imaginary
        stoichiometry = [] if stoichiometry is None else stoichiometry.values
        # find the longest string to figure out padding
        max_fnm = max(np.vectorize(len)(filenames).max(), 20)
        try:
            max_stoich = max(np.vectorize(len)(stoichiometry).max(), 13)
        except ValueError:
            max_stoich = 0
        # values should be in shape (file, genre)
        values = np.array([en.values for en in energies]).T
        # populations in percent
        popul = np.array([en.populations * 100 for en in energies]).T
        names = [self._header[en.genre] for en in energies]
        population_widths = [max(8, len(n)) for n in names]
        population_header = "  ".join(
            [f"{n:<{w}}" for n, w in zip(names, population_widths)]
        )
        energies_widths = [14 if n == "SCF" else 12 for n in names]
        energies_header = "  ".join(
            [f"{n:<{w}}" for n, w in zip(names, energies_widths)]
        )
        names_line = [" " * max_fnm, population_header, energies_header]
        names_line += ["    "] if frequencies is not None else []
        names_line += ["             "] if max_stoich else []
        names_line = " | ".join(names_line)
        precisions = [8 if n == "SCF" else 6 for n in names]
        header = [
            f"{'Gaussian output file':<{max_fnm}}",
            f"{'Population / %':^{len(population_header)}}",
            f"{'Energy / Hartree':^{len(energies_header)}}",
        ]
        header += ["Imag"] if frequencies is not None else []
        header += [f"{'Stoichiometry':<{max_stoich}}"] if max_stoich else []
        header = " | ".join(header)
        line_format = (
            f"{{:<{max_fnm}}} | {{}} | {{}}"
            f"{' | {:^ 4}' if frequencies is not None else '{}'}"
            f"{f' | {{:<{max_stoich}}}' if max_stoich else '{}'}\n"
        )
        template_params = {"cat": "overview", "conf": "multiple"}
        with self._get_handle(name_template, template_params) as file:
            file.write(header + "\n")
            file.write(names_line + "\n")
            file.write("-" * len(header) + "\n")
            rows = zip_longest(
                filenames, values, popul, imaginary, stoichiometry, fillvalue=""
            )
            for fnm, vals, pops, imag, stoich in rows:
                p_line = "  ".join(
                    [f"{p:>{w}.4f}" for p, w in zip(pops, population_widths)]
                )
                v_line = "  ".join(
                    [
                        f"{v:> {w}.{p}f}"
                        for v, w, p in zip(vals, energies_widths, precisions)
                    ]
                )
                line = line_format.format(fnm, p_line, v_line, imag, stoich)
                file.write(line)
        logger.info("Energies collective export to text file done.")

    def energies(
        self,
        energies: Energies,
        corrections: Optional[FloatArray] = None,
        name_template: Union[str, Template] = "distribution-${genre}.${ext}",
    ):
        """Writes Energies object to txt file.

        Parameters
        ----------
        energies: glassware.Energies
            Energies object that is to be serialized
        corrections: glassware.DataArray, optional
            DataArray object, containing energies corrections
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.
        """
        max_fnm = max(np.vectorize(len)(energies.filenames).max(), 20)
        header = [f"{'Gaussian output file':<{max_fnm}}"]
        header += ["Population/%", "Min.B.Factor", "DE/(kcal/mol)", "Energy/Hartree"]
        header += ["Corr/Hartree"] if corrections is not None else []
        header = " | ".join(header)
        align = ("<", ">", ">", ">", ">", ">")
        width = (max_fnm, 12, 12, 13, 14, 12)
        corrections = corrections.values if corrections is not None else []
        fmt = (
            "",
            ".4f",
            ".4f",
            ".4f",
            ".8f" if energies.genre == "scf" else ".6f",
            "f",
        )
        rows = zip_longest(
            energies.filenames,
            energies.populations * 100,
            energies.min_factors,
            energies.deltas,
            energies.values,
            corrections,
            fillvalue=None,
        )
        template_params = {
            "conf": "multiple",
            "genre": energies.genre,
            "cat": "populations",
        }
        with self._get_handle(name_template, template_params) as file:
            file.write(header + "\n")
            file.write("-" * len(header) + "\n")
            for row in rows:
                new_row = [
                    f"{v:{a}{w}{f}}"
                    for v, a, w, f in zip(row, align, width, fmt)
                    if v is not None
                ]
                file.write(" | ".join(new_row) + "\n")
        logger.info("Energies separate export to text files done.")

    def single_spectrum(
        self,
        spectrum: SingleSpectrum,
        name_template: Union[str, Template] = "${cat}.${genre}-${det}.${ext}",
    ):
        """Writes SingleSpectrum object to txt file.

        Parameters
        ----------
        spectrum: glassware.SingleSpectrum
            spectrum, that is to be serialized
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.
        """
        title = (
            f"{spectrum.genre} calculated with peak width = "
            f'{spectrum.width} {spectrum.units["width"]} and '
            f'{spectrum.fitting} fitting, shown as {spectrum.units["x"]} '
            f'vs. {spectrum.units["y"]}'
        )
        template_params = {
            "genre": spectrum.genre,
            "cat": "spectrum",
            "det": spectrum.averaged_by,
        }
        with self._get_handle(name_template, template_params) as file:
            file.write(title + "\n")
            if spectrum.averaged_by:
                file.write(
                    f"{len(spectrum.filenames)} conformers averaged base on"
                    f" {self._header[spectrum.averaged_by]}\n"
                )
            file.write(
                "\n".join(
                    f"{int(x):> 7.2f}\t{y: > 10.4f}"
                    for x, y in zip(spectrum.x, spectrum.y)
                )
            )
        logger.info("Spectrum export to text files done.")

    def spectral_activities(
        self,
        band: SpectralActivities,
        data: List[SpectralActivities],
        name_template: Union[str, Template] = "${conf}.${cat}-${det}.${ext}",
    ):
        """Writes SpectralActivities objects to txt files (one for each conformer).

        Parameters
        ----------
        band: glassware.SpectralActivities
            object containing information about band at which transitions occur;
            it should be frequencies for vibrational data and wavelengths or
            excitation energies for electronic data
        data: list of glassware.SpectralActivities
            SpectralActivities objects that are to be serialized; all should contain
            information for the same conformers. Assumes that all *data*'s elements have
            the same *spectra_type*, which is passed to the *name_template* as "det".
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.

        Raises
        ------
        ValueError
            if *data* is an empty sequence
        """
        self._spectral(
            band=band,
            data=data,
            name_template=name_template,
            category="activities",
        )

    def spectral_data(
        self,
        band: SpectralActivities,
        data: List[SpectralData],
        name_template: Union[str, Template] = "${conf}.${cat}-${det}.${ext}",
    ):
        """Writes SpectralData objects to txt files (one for each conformer).

        Parameters
        ----------
        band: glassware.SpectralData
            object containing information about band at which transitions occur;
            it should be frequencies for vibrational data and wavelengths or
            excitation energies for electronic data
        data: list of glassware.SpectralData
            SpectralData objects that are to be serialized; all should contain
            information for the same conformers. Assumes that all *data*'s elements have
            the same *spectra_type*, which is passed to the *name_template* as "det".
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.

        Raises
        ------
        ValueError
            if *data* is an empty sequence
        """
        self._spectral(
            band=band, data=data, name_template=name_template, category="data"
        )

    def _spectral(
        self,
        band: SpectralActivities,
        data: Union[List[SpectralData], List[SpectralActivities]],
        name_template: Union[str, Template],
        category: str,
    ):
        """Writes SpectralData objects to txt files (one for each conformer).

        Parameters
        ----------
        band: glassware.SpectralData
            object containing information about band at which transitions occur;
            it should be frequencies for vibrational data and wavelengths or
            excitation energies for electronic data
        data: list of glassware.SpectralData
            SpectralData objects that are to be serialized; all should contain
            information for the same conformers. Assumes that all *data*'s elements have
            the same *spectra_type*, which is passed to the *name_template* as "det".
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.
        category : str
            category of exported data genres

        Raises
        ------
        ValueError
            if *data* is an empty sequence
        """
        try:
            spectra_type = data[0].spectra_type
        except IndexError:
            raise ValueError("No data to export.")
        data = [band] + data
        genres = [bar.genre for bar in data]
        headers = [self._header[genre] for genre in genres]
        widths = [self._formatters[genre][4:-4] for genre in genres]
        formatted = [f"{h: <{w}}" for h, w in zip(headers, widths)]
        values = zip(*[bar.values for bar in data])
        template_params = {"genre": band.genre, "cat": category, "det": spectra_type}
        for values_, handle in zip(
            values,
            self._iter_handles(band.filenames, name_template, template_params),
        ):
            handle.write("\t".join(formatted))
            handle.write("\n")
            for vals in zip(*values_):
                line = "\t".join(
                    self._formatters[g].format(v) for v, g in zip(vals, genres)
                )
                handle.write(line + "\n")
        logger.info(f"{category.title()} export to text files done.")

    def spectra(
        self,
        spectra: Spectra,
        name_template: Union[str, Template] = "${conf}.${genre}.${ext}",
    ):
        """Writes Spectra object to text files (one for each conformer).

        Parameters
        ----------
        spectra: glassware.Spectra
            Spectra object, that is to be serialized
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.
        """
        abscissa = spectra.x
        title = (
            f"{spectra.genre} calculated with peak width = {spectra.width}"
            f' {spectra.units["width"]} and {spectra.fitting} '
            f'fitting, shown as {spectra.units["x"]} vs. '
            f'{spectra.units["y"]}'
        )
        template_params = {
            "genre": spectra.genre,
            "cat": "spectra",
            "det": spectra.spectra_type,
        }
        abscissa_genre = "wavelen" if spectra.spectra_type == "electronic" else "freq"
        values_template = "\t".join(
            (self._formatters[abscissa_genre], self._formatters[spectra.genre])
        )
        for values, handle in zip(
            spectra.y,
            self._iter_handles(spectra.filenames, name_template, template_params),
        ):
            handle.write(title + "\n")
            handle.write(
                "\n".join(
                    values_template.format(a, v) for a, v in zip(abscissa, values)
                )
            )
        logger.info("Spectra export to text files done.")

    def transitions(
        self,
        transitions: Transitions,
        wavelengths: Bands,
        only_highest=True,
        name_template: Union[str, Template] = "${conf}.${cat}-${det}.${ext}",
    ):
        """Writes electronic transitions data to text files (one for each conformer).

        Parameters
        ----------
        transitions : glassware.Transitions
            Electronic transitions data that should be serialized.
        wavelengths : glassware.ElectronicActivities
            Object containing information about wavelength at which transitions occur.
        only_highest : bool
            Specifies if only transition of highest contribution to given band should
            be reported. If ``False`` all transition are saved to file.
            Defaults to ``True``.
        name_template : str or string.Template
            Template that will be used to generate filenames. Refer to
            :meth:`.make_name` documentation for details on supported placeholders.

        """
        transtions_data = (
            transitions.highest_contribution
            if only_highest
            else (
                transitions.ground,
                transitions.excited,
                transitions.values,
                transitions.contribution,
            )
        )
        contrib_text = "of highest contribution" if only_highest else "contributing"
        title = f"Electronic transitions {contrib_text} to each band."
        legend = "wavelength: ground -> excited, coefficient, contribution"
        transition_entry = "{:>6d} -> {:<7d} {:> 11.5f} {:>12.0%}\n"
        template_params = {
            "genre": transitions.genre,
            "cat": "transitions",
            "det": "highest" if only_highest else "all",
        }
        for grounds, exciteds, values, contribs, bands, handle in zip(
            *transtions_data,
            wavelengths.wavelen,
            self._iter_handles(transitions.filenames, name_template, template_params),
        ):
            handle.write(title + "\n")
            handle.write(legend + "\n\n")
            for g, e, v, c, b in zip(grounds, exciteds, values, contribs, bands):
                try:
                    listed = [
                        transition_entry.format(*d)
                        for d in zip(g, e, v, c)
                        # omit entry if any value is masked
                        if all(x is not np.ma.masked for x in d)
                    ]
                    listed = (" " * 12).join(listed)
                except TypeError:
                    # transition_data is transitions.highest_contribution
                    listed = transition_entry.format(g, e, v, c)
                handle.write(f"  {b:> 7.2f} nm: {listed}")



"""Parser for reading spectra calculation parameters from file."""
import logging
import re
from configparser import ConfigParser, MissingSectionHeaderError, ParsingError
from pathlib import Path
from typing import Callable, Union

from .. import datawork as dw
from .parser_base import _PARSERS

logger = logging.getLogger(__name__)

# from Python re docs
FLOATCRE = re.compile(r"[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?")
FITTINGCRE = re.compile(r"lorentzian|gaussian")


def quantity(s: str) -> float:
    """Convert to float first occurrence of float-looking part of string *s*, ignoring
    anything else. Raise :class:`configparser.ParsingError` if float cannot be found.

    Parameters
    ----------
    s : str
        string containing a float

    Returns
    -------
    float
        extracted float value
    """
    match = FLOATCRE.search(s)
    if match:
        return float(match.group())
    else:
        raise ParsingError("Cannot interpret any part of given string as float.")


def fitting(s: str) -> Callable:
    """Get fitting function mentioned in a given string *s*, ignoring anything else.
    Raise :class:`configparser.ParsingError` if known function name cannot be found.

    Parameters
    ----------
    s : str
        string containing name of fitting function

    Returns
    -------
    callable
        an identified fitting function
    """
    match = FITTINGCRE.search(s.lower())
    try:
        return getattr(dw, match.group())
    except AttributeError:
        raise ParsingError(f"No such fitting function: {s}.")


class ParametersParser:
    """Parser for configuration files containing spectra calculation parameters.

    Configuration file should be in format similar to .ini files: a list of key-value
    pairs, separated with "=" or ":", one pair for line. Standard parameters (width,
    start, stop, step, and fitting) will be converted to appropriate data type, i.e.
    float or function reference. If parameter value cannot be converted to its
    target type, it will be ignored and warning will be emitted. Any other (unexpected)
    parameters are included in the output and left as they are.

    The parser is case-insensitive and knows some alias names of expected parameters:
    for instance, "hwhm", "half width of band in half height", "half width at half
    maximum" will be all recognized as "width" parameter. If you wish to add custom
    aliases, update :attr:`ParametersParser.ALIASES` dictionary with appropriate
    "alias": "target" pair.

    Notes
    -----
    :class:`ParametersParser` is using Python's configparser, so it will parse files
    that contain a section header, enclosed in braces. However, the section name will be
    ignored and there may be only one such section, otherwise an exception is raised.
    """

    purpose = "parameters"
    extensions = ("txt", "cfg", "ini")

    ALIASES = {
        "half width of band in half height": "width",
        "half width at half maximum": "width",
        "hwhm": "width",
        "start range": "start",
        "stop range": "stop",
        "fitting function": "fitting",
    }

    def __init__(self):
        self._parser = ConfigParser(
            converters={"quantity": quantity, "fitting": fitting}
        )
        self._parser.optionxform = self.optionxform
        self._transformers = {
            "width": self._parser.getquantity,
            "start": self._parser.getquantity,
            "stop": self._parser.getquantity,
            "step": self._parser.getquantity,
            "fitting": self._parser.getfitting,
        }

    def optionxform(self, optionstr: str) -> str:
        """Translates option names to desired form - lowercase and standard wording,
        as defined in ALIASES.
        """
        optionstr = optionstr.lower()
        if optionstr in self.ALIASES:
            optionstr = self.ALIASES[optionstr]
        return optionstr

    @property
    def parameters(self) -> dict:
        """Dictionary of parameters for calculating spectra extracted from parsed file
        and converted to appropriate type."""
        params = {}
        try:
            section = self._parser.sections()[0]
        except IndexError:
            return params
        for option in self._parser.options(section):
            try:
                params[option] = self._transformers[option](section, option)
            except ParsingError:
                logger.warning(
                    f"Cannot understand value for {option}, it will be ignored."
                )
            except KeyError:
                logger.warning(
                    f"Unknown parameter in settings: '{option}',"
                    f" it will not be converted."
                )
                params[option] = self._parser.get(section, option)
        return params

    def parse(self, source: Union[str, Path]):
        """Parse given source file to get stored parameters.

        Parameters
        ----------
        source : str or Path
            Path to file with calculations' parameters.

        Returns
        -------
        dict
            Parsed parameters.
        """
        source = Path(source)
        with source.open() as handle:
            text = handle.read()
        try:
            self._parser.read_string(text)
        except MissingSectionHeaderError:
            self._parser.read_string(f"[PARAMETERS]\n{text}")
        if len(self._parser.sections()) > 1:
            raise ParsingError(
                "Multiple sections in parameters setup file are not supported."
            )
        return self.parameters


_PARSERS[ParametersParser.purpose] = ParametersParser



"""All core algorithms for data handling live here, along with some helpers.

This package implements functions used by ``tesliper`` to provide its core
functionality: comparing conformers, calculating spectra, averaging them, etc. It is
divided into few modules based on a field that given function is related to.
"""

from .atoms import Atom, atomic_number, symbol_of_element, validate_atoms
from .energies import (
    BOLTZMANN,
    calculate_deltas,
    calculate_min_factors,
    calculate_populations,
)
from .geometry import (
    calc_rmsd,
    center,
    drop_atoms,
    fixed_windows,
    get_triangular,
    get_triangular_base,
    is_triangular,
    kabsch_rotate,
    pyramid_windows,
    rmsd_sieve,
    stretching_windows,
    take_atoms,
)
from .intensities import (
    DEFAULT_ACTIVITIES,
    dip_to_ir,
    dip_to_uv,
    osc_to_uv,
    rot_to_ecd,
    rot_to_vcd,
)
from .spectra import (
    calculate_average,
    calculate_spectra,
    convert_band,
    count_imaginary,
    find_imaginary,
    find_offset,
    find_scaling,
    gaussian,
    idx_offset,
    lorentzian,
    unify_abscissa,
)



from pathlib import Path
from unittest import mock

import pytest
from hypothesis import assume, given
from hypothesis import strategies as st

from tesliper import Energies, InfoArray, Spectra
from tesliper.glassware import (
    Bands,
    ElectronicActivities,
    ElectronicData,
    FloatArray,
    ScatteringActivities,
    ScatteringData,
    SingleSpectrum,
    VibrationalActivities,
    VibrationalData,
)
from tesliper.writing import WriterBase


@pytest.fixture(scope="session")
def writer_class():
    class _Writer(WriterBase):
        extension = "txt"

    return _Writer


@pytest.fixture
def path_is_dir(monkeypatch):
    monkeypatch.setattr(Path, "is_dir", lambda _: True)


@pytest.fixture
def path_is_not_dir(monkeypatch):
    monkeypatch.setattr(Path, "is_dir", lambda _: False)


@pytest.fixture
def path_is_file(monkeypatch):
    monkeypatch.setattr(Path, "is_file", lambda _: True)


@pytest.fixture
def path_is_not_file(monkeypatch):
    monkeypatch.setattr(Path, "is_file", lambda _: False)


@pytest.fixture
def path_exists(monkeypatch):
    monkeypatch.setattr(Path, "exists", lambda _: True)


@pytest.fixture
def path_doesnt_exists(monkeypatch):
    monkeypatch.setattr(Path, "exists", lambda _: False)


@pytest.mark.usefixtures("path_is_not_file", "path_is_not_dir", "path_doesnt_exist")
@pytest.fixture
def path_truly_doesnt_exist():
    yield


@pytest.fixture
def no_parent_dir(monkeypatch):
    parent = mock.Mock()
    parent.exists = mock.Mock(return_value=False)
    monkeypatch.setattr(Path, "parent", parent)


@given(mode=st.text())
def test_writer_mode_init(writer_class, mode):
    assume(mode not in ("a", "x", "w"))
    with pytest.raises(ValueError):
        writer_class(destination="", mode=mode)


@pytest.mark.usefixtures("path_exists")
@pytest.mark.parametrize("mode", ["a", "x", "w"])
def test_writer_mode_ok_init(writer_class, mode):
    wrt = writer_class(destination="", mode=mode)
    assert wrt.mode == mode


@pytest.mark.usefixtures("path_is_not_dir")
@pytest.mark.parametrize("mode", ["a", "x", "w"])
def test_writer_init_dest_is_not_dir(writer_class, mode):
    with pytest.raises(FileNotFoundError):
        writer_class(destination="", mode=mode)


@pytest.mark.usefixtures("path_doesnt_exists")
def test_writer_check_file_no_dest(writer_class, monkeypatch):
    parent_mock = mock.Mock()
    parent_mock.exists = mock.Mock(return_value=True)
    monkeypatch.setattr(Path, "parent", parent_mock)
    wrt = writer_class(destination="", mode="w")
    assert wrt.check_file("") == Path("")


@pytest.mark.usefixtures("path_exists")
def test_writer_check_file_overwrite(writer_class):
    wrt = writer_class(destination="", mode="w")
    assert wrt.check_file("") == Path("")


@pytest.mark.usefixtures("path_doesnt_exists")
@pytest.mark.usefixtures("no_parent_dir")
def test_writer_check_file_no_parent_dir(writer_class):
    wrt = writer_class(destination="", mode="w")
    with pytest.raises(FileNotFoundError):
        wrt.check_file("")


@pytest.mark.usefixtures("path_exists")
def test_writer_check_file_new_only_no_dest(writer_class):
    wrt = writer_class(destination="", mode="x")
    with pytest.raises(FileExistsError):
        wrt.check_file("")


arrays_by_type = dict(
    energies=Energies("gib", [""], [1]),
    vibrationaldata=VibrationalData("emang", [""], [[1]], [[1]]),
    scatteringdata=ScatteringData("depp", [""], [[1]], [[1]]),
    electronicdata=ElectronicData("eemang", [""], [[1]], [[1]]),
    vibrationalactivities=VibrationalActivities("iri", [""], [[1]], [[1]]),
    scatteringactivities=ScatteringActivities("ramact", [""], [[1]], [[1]]),
    electronicactivities=ElectronicActivities("vrot", [""], [[1]], [[1]]),
    spectra=Spectra("ir", [""], [[1, 2]], [1, 2]),
    singlespectrum=SingleSpectrum("ir", [1, 2], [1, 2]),
    infoarray=InfoArray("command", [""], [""]),
)
extras_by_type = dict(
    corrections=FloatArray("gibcorr", [""], [1]),
    frequencies=Bands("freq", [""], [[1]], [[1]]),
    wavelengths=Bands("wavelen", [""], [[1]], [[1]]),
    stoichiometry=InfoArray("stoichiometry", [""], [""]),
)


@pytest.mark.parametrize("arraytype,array", tuple(arrays_by_type.items()))
def test_distribution_data(writer_class, arraytype, array):
    output, extras = writer_class.distribute_data([array])
    assert isinstance(output, dict)
    assert arraytype in output
    assert output[arraytype] is not None


@pytest.mark.parametrize("arraytype,array", tuple(extras_by_type.items()))
def test_distribution_extras(writer_class, arraytype, array):
    output, extras = writer_class.distribute_data([array])
    assert isinstance(extras, dict)
    assert arraytype in extras
    assert extras[arraytype] is not None


writer_methods = {
    "generic": dict(data=None),
    "overview": dict(
        energies=None,
        frequencies=None,
        stoichiometry=None,
    ),
    "energies": dict(energies=None, corrections=None),
    "single_spectrum": dict(spectrum=None),
    "spectral_activities": dict(band=None, data=None),
    "spectra": dict(spectra=None),
    "transitions": dict(
        transitions=None,
        wavelengths=None,
        only_highest=True,
    ),
    "geometry": dict(
        geometry=None,
        charge=None,
        multiplicity=None,
    ),
}


@pytest.mark.parametrize("method,signature", tuple(writer_methods.items()))
def test_not_implemented_methods(writer_class, method, signature):
    wrt = writer_class(destination="")
    m = getattr(wrt, method)
    with pytest.raises(NotImplementedError):
        m(**signature)



from itertools import chain
from pathlib import Path

import numpy as np
import pytest
from hypothesis import given
from hypothesis import strategies as st

from tesliper import Tesliper

fixtures_dir = Path(__file__).parent / "fixtures"


@pytest.fixture
def wanted_files():
    return ["meoh-1.out", "meoh-2.out"]


@pytest.fixture
def tesliper(tmp_path, wanted_files):
    return Tesliper(
        input_dir=fixtures_dir, output_dir=tmp_path, wanted_files=wanted_files
    )


@pytest.fixture
def inconsistent(tesliper):
    tesliper.allow_data_inconsistency = True
    return tesliper


@pytest.fixture
def extracted(tesliper):
    tesliper.extract()
    return tesliper


@pytest.fixture
def trimmed(extracted):
    extracted.conformers.kept = [False, True]
    return extracted


@pytest.fixture
def calculated(extracted):
    extracted.calculate_spectra(["iri"])
    return extracted


@pytest.fixture
def averaged(calculated):
    calculated.average_spectra()
    return calculated


def resurect(tesliper, path):
    file = path / ".tslr"
    tesliper.serialize(file)
    return Tesliper.load(file)


def test_serialization_init_params(tesliper, tmp_path):
    resurrected = resurect(tesliper, tmp_path)
    assert resurrected.wanted_files == tesliper.wanted_files
    assert resurrected.input_dir == tesliper.input_dir
    assert resurrected.output_dir == tesliper.output_dir


def test_serialization_calc_params(tesliper, tmp_path):
    resurrected = resurect(tesliper, tmp_path)
    old_params = {key: params.copy() for key, params in tesliper.parameters.items()}
    new_params = {key: params.copy() for key, params in resurrected.parameters.items()}
    for params in chain(old_params.values(), new_params.values()):
        params["fitting"] = params["fitting"].__name__
    assert new_params == old_params


def test_serialization_conformers(extracted, tmp_path):
    resurrected = resurect(extracted, tmp_path)
    assert resurrected.conformers == extracted.conformers


def test_serialization_trimmed(trimmed, tmp_path):
    resurrected = resurect(trimmed, tmp_path)
    assert resurrected.conformers.kept == trimmed.conformers.kept


def test_serialization_inconsistent(inconsistent, tmp_path):
    resurrected = resurect(inconsistent, tmp_path)
    assert (
        resurrected.conformers.allow_data_inconsistency
        == inconsistent.conformers.allow_data_inconsistency
    )


def assert_spectra_equal(new, old):
    for key, value in new.__dict__.items():
        if not isinstance(value, np.ndarray):
            assert value == getattr(old, key)
        else:
            assert np.array(value == getattr(old, key)).all()


def test_serialization_spectra(calculated, tmp_path):
    resurrected = resurect(calculated, tmp_path)
    for genre, spc in resurrected.spectra.items():
        assert_spectra_equal(spc, calculated.spectra[genre])


def test_serialization_averaged(averaged, tmp_path):
    resurrected = resurect(averaged, tmp_path)
    for key, spc in resurrected.averaged.items():
        assert_spectra_equal(spc, averaged.averaged[key])


@pytest.mark.xfail(reason="Not implemented yet.")
def test_serialization_experimental(experimental, tmp_path):
    resurrected = resurect(experimental, tmp_path)
    assert resurrected.experimental == experimental.experimental


def test_export_data(averaged, wanted_files):
    averaged.export_data(["iri", "scf"])
    files = list(averaged.output_dir.iterdir())
    expected = {
        Path(f).with_suffix(".activities-vibrational").name for f in wanted_files
    }
    expected.update({"overview", "distribution-scf"})
    assert {f.stem for f in files} == expected


def test_export_data_empty(averaged):
    averaged.export_data(["rot", "gib"])
    files = list(averaged.output_dir.iterdir())
    expected = {"overview", "distribution-gib"}
    assert {f.stem for f in files} == expected


@pytest.mark.parametrize("fmt", ["txt", "csv", "xlsx"])
def test_export_energies(extracted, fmt):
    extracted.export_energies(fmt=fmt)
    files = list(extracted.output_dir.iterdir())
    expected = {f"distribution-{e}" for e in ("gib", "ent", "zpe", "ten", "scf")}
    if fmt == "xlsx":
        assert len(files) == 1
        assert files[0].suffix == ".xlsx"
    elif fmt == "txt":
        expected.add("overview")
        assert {f.stem for f in files} == expected
    elif fmt == "csv":
        assert {f.stem for f in files} == expected


@pytest.mark.parametrize("fmt", ["txt", "csv", "xlsx"])
def test_export_spectral_data(extracted, fmt, wanted_files):
    extracted.export_spectral_data(fmt=fmt)
    files = list(extracted.output_dir.iterdir())
    if fmt == "xlsx":
        assert len(files) == 1
        assert files[0].suffix == ".xlsx"
    elif fmt == "txt":
        expected = {
            Path(f).with_suffix(f".data-{t}").name
            for f in wanted_files
            for t in ("vibrational", "scattering")
        }
        assert {f.stem for f in files} == expected


@pytest.mark.parametrize("fmt", ["txt", "csv", "xlsx"])
def test_export_activities(extracted, fmt, wanted_files):
    extracted.export_activities(fmt=fmt)
    files = list(extracted.output_dir.iterdir())
    if fmt == "xlsx":
        assert len(files) == 1
        assert files[0].suffix == ".xlsx"
    else:
        expected = {
            Path(f).with_suffix(f".activities-{t}").name
            for f in wanted_files
            for t in ("vibrational", "scattering")
        }
        assert {f.stem for f in files} == expected


@pytest.mark.parametrize("fmt", ["txt", "csv", "xlsx"])
def test_export_spectra(calculated, fmt, wanted_files):
    calculated.export_spectra(fmt=fmt)
    files = list(calculated.output_dir.iterdir())
    if fmt == "xlsx":
        assert len(files) == 1
        assert files[0].suffix == ".xlsx"
    else:
        assert {f.stem for f in files} == {
            Path(f).with_suffix(".ir").name for f in wanted_files
        }


@pytest.mark.parametrize("fmt", ["txt", "csv", "xlsx"])
def test_export_averaged(averaged, fmt):
    averaged.export_averaged(fmt=fmt)
    files = list(averaged.output_dir.iterdir())
    if fmt == "xlsx":
        assert len(files) == 1
        assert files[0].suffix == ".xlsx"
    else:
        assert {f.stem for f in files} == {
            f"spectrum.ir-{e}" for e in ("gib", "ent", "zpe", "ten", "scf")
        }


def test_export_job_file(extracted, wanted_files):
    extracted.export_job_file()
    files = list(extracted.output_dir.iterdir())
    assert {f.name for f in files} == {
        Path(f).with_suffix(".gjf").name for f in wanted_files
    }


@given(value=st.floats(min_value=0, allow_nan=False, exclude_min=True))
def test_subscription_temperature(value):
    tslr = Tesliper()
    tslr.temperature = value
    array = tslr["zpe"]
    assert array.t == value


def test_get_averaged_spectrum(calculated):
    spc = calculated.get_averaged_spectrum("ir", "zpe")
    assert spc.genre == "ir"
    assert spc.averaged_by == "zpe"


def test_get_averaged_calculate(calculated, extracted):
    e = extracted.get_averaged_spectrum("ir", "zpe")
    c = calculated.get_averaged_spectrum("ir", "zpe")
    assert e.values.tolist() == c.values.tolist()


def test_get_average_raises_missing(tesliper):
    assert "ir" not in tesliper.spectra
    assert not tesliper.conformers.has_genre("dip")
    with pytest.raises(ValueError):
        tesliper.get_averaged_spectrum("ir", "zpe")



import os
from unittest.mock import patch

import pytest

from tesliper.extraction import gaussian_parser as gprs

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture(scope="class")
def setup_class(request):
    patchers = []
    parser = gprs.GaussianParser()
    for name in dir(parser):
        method = getattr(parser, name)
        if hasattr(method, "is_state"):
            patcher = patch.object(parser, name, wraps=method)
            mock = patcher.start()
            # need to manually patch the reference in parser.states dictionary
            parser.states[name] = mock
            patchers.append(patcher)
    try:
        request.cls.parser = parser
        path = os.path.join(THIS_DIR, os.pardir, "fixtures", request.cls.file)
        with open(path, "r") as file:
            request.cls.data = parser.parse(file)
    except Exception:
        for patcher in patchers:
            patcher.stop()
            parser.states[patcher.attribute] = getattr(parser, patcher.attribute)
        raise


class ParserTestCase:
    """This is mix-in class for test cases for gaussian files parsing."""

    # name of gaussian output file in fixtures dir
    file = ""  # molecule is formaldehyde by default
    data = {}  # dictionary populated by 'setup_class' fixture

    def test_normal_termination(self):
        assert self.data["normal_termination"]

    def test_version(self):
        assert self.data["version"] == "Gaussian 09, Revision E.01"

    def test_charge(self):
        assert self.data["charge"] == 0.0

    def test_multiplicity(self):
        assert self.data["multiplicity"] == 1.0

    def test_stoichiometry(self):
        assert self.data["stoichiometry"] == "CH2O"

    def test_atoms(self):
        assert self.data["last_read_atoms"] == [6, 1, 1, 8]

    def test_input_atoms(self):
        assert self.data["input_atoms"] == ["C", "H", "H", "O"]


@pytest.mark.usefixtures("setup_class")
class TestFreq(ParserTestCase):
    file = "fal-freq.out"

    @pytest.mark.skip("Doesn't work as expected, see GaussianParser.optimization")
    def test_methods_called(self):
        self.parser.initial.assert_called()
        self.parser.wait.assert_called()
        self.parser.frequencies.assert_called()
        self.parser.excited.assert_not_called()
        self.parser.optimization.assert_not_called()
        self.parser.geometry.assert_called()

    def test_keys(self):
        assert set(self.data.keys()) == {
            "normal_termination",
            "version",
            "command",
            "charge",
            "multiplicity",
            "input_geom",
            "input_atoms",
            "scf",
            "stoichiometry",
            "zpecorr",
            "tencorr",
            "entcorr",
            "gibcorr",
            "zpe",
            "ten",
            "ent",
            "gib",
            "freq",
            "mass",
            "frc",
            "iri",
            "depolarp",
            "depolaru",
            "ramanactiv",
            "last_read_geom",
            "last_read_atoms",
        }

    def test_command(self):
        assert self.data["command"] == "freq hf/3-21g"

    def test_scf(self):
        assert self.data["scf"] == -113.217409254

    def test_zpecorr(self):
        assert self.data["zpecorr"] == 0.028669

    def test_tencorr(self):
        assert self.data["tencorr"] == 0.031529

    def test_entcorr(self):
        assert self.data["entcorr"] == 0.032474

    def test_gibcorr(self):
        assert self.data["gibcorr"] == 0.006971

    def test_zpe(self):
        assert self.data["zpe"] == -113.188740

    def test_ten(self):
        assert self.data["ten"] == -113.185880

    def test_ent(self):
        assert self.data["ent"] == -113.184936

    def test_gib(self):
        assert self.data["gib"] == -113.210438

    def test_freq(self):
        assert self.data["freq"] == [
            1279.9047,
            1337.9260,
            1548.4195,
            1717.9022,
            3288.8988,
            3411.2493,
        ]

    def test_mass(self):
        assert self.data["mass"] == [1.3348, 1.3041, 2.9470, 1.4267, 1.0451, 1.1203]

    def test_frc(self):
        assert self.data["frc"] == [1.2883, 1.3753, 4.1630, 2.4808, 6.6606, 7.6808]

    def test_iri(self):
        assert self.data["iri"] == [11.0336, 11.8913, 50.4624, 10.5434, 9.5817, 67.3121]

    def test_depolarp(self):
        assert self.data["depolarp"] == [0.7500, 0.7499, 0.4706, 0.2785, 0.1912, 0.7500]

    def test_depolaru(self):
        assert self.data["depolaru"] == [0.8571, 0.8571, 0.6400, 0.4357, 0.3211, 0.8571]

    def test_ramanactiv(self):
        assert self.data["ramanactiv"] == [
            0.0218,
            6.0244,
            21.4502,
            1.3596,
            110.6841,
            38.7779,
        ]

    def test_geometry(self):
        assert self.data["last_read_geom"] == [
            [0.000199, 0.562211, 0.000000],
            [0.924720, 1.100877, 0.000000],
            [-0.927506, 1.095374, 0.000000],
            [0.000199, -0.696189, 0.000000],
        ]


@pytest.mark.usefixtures("setup_class")
class TestFreqRoa(ParserTestCase):
    file = "fal-freq-roa.out"

    @pytest.mark.skip("Doesn't work as expected, see GaussianParser.optimization")
    def test_methods_called(self):
        self.parser.initial.assert_called()
        self.parser.wait.assert_called()
        self.parser.frequencies.assert_called()
        self.parser.excited.assert_not_called()
        self.parser.optimization.assert_not_called()
        self.parser.geometry.assert_called()

    def test_keys(self):
        assert set(self.data.keys()) == {
            "normal_termination",
            "version",
            "command",
            "charge",
            "multiplicity",
            "input_geom",
            "input_atoms",
            "scf",
            "stoichiometry",
            "last_read_geom",
            "last_read_atoms",
            "zpecorr",
            "tencorr",
            "entcorr",
            "gibcorr",
            "zpe",
            "ten",
            "ent",
            "gib",
            "freq",
            "mass",
            "frc",
            "iri",
            "ramact",
            "depp",
            "depu",
            "alpha2",
            "beta2",
            "alphag",
            "gamma2",
            "delta2",
            "raman1",
            "roa1",
            "cid1",
            "raman2",
            "roa2",
            "cid2",
            "raman3",
            "roa3",
            "cid3",
            "rc180",
        }

    def test_command(self):
        assert self.data["command"] == "freq=roa hf/3-21g"

    def test_freq(self):
        assert self.data["freq"] == [
            1279.9047,
            1337.9260,
            1548.4195,
            1717.9022,
            3288.8988,
            3411.2493,
        ]

    def test_mass(self):
        assert self.data["mass"] == [1.3348, 1.3041, 2.9470, 1.4267, 1.0451, 1.1203]

    def test_frc(self):
        assert self.data["frc"] == [1.2883, 1.3753, 4.1630, 2.4808, 6.6606, 7.6808]

    def test_iri(self):
        assert self.data["iri"] == [11.0336, 11.8913, 50.4623, 10.5434, 9.5818, 67.3121]

    def test_ramact(self):
        assert self.data["ramact"] == [
            0.0681,
            6.1822,
            25.3142,
            1.4865,
            119.9701,
            41.7021,
        ]

    def test_depp(self):
        assert self.data["depp"] == [0.7500, 0.7499, 0.4633, 0.1772, 0.1971, 0.7500]

    def test_depu(self):
        assert self.data["depu"] == [0.8571, 0.8571, 0.6332, 0.3010, 0.3293, 0.8571]

    def test_alpha2(self):
        assert self.data["alpha2"] == [0.0000, 0.0000, 0.1470, 0.0214, 1.6417, 0.0000]

    def test_beta2(self):
        assert self.data["beta2"] == [0.0097, 0.8831, 2.6716, 0.0746, 6.5849, 5.9574]

    def test_alphag(self):
        assert self.data["alphag"] == [0.0000, -0.0000, -0.0000, 0.0000, 0.0000, 0.0000]

    def test_gamma2(self):
        assert self.data["gamma2"] == [0.0000, -0.0000, -0.0000, 0.0000, 0.0000, 0.0000]

    def test_delta2(self):
        assert self.data["delta2"] == [
            -0.0000,
            0.0000,
            -0.0000,
            -0.0000,
            -0.0000,
            -0.0000,
        ]

    def test_raman1(self):
        assert self.data["raman1"] == [0.272, 24.729, 101.257, 5.946, 479.880, 166.808]

    def test_roa1(self):
        assert self.data["roa1"] == [0.000, -0.000, -0.000, 0.000, 0.000, 0.000]

    def test_cid1(self):
        assert self.data["cid1"] == [0.000, -0.000, -0.000, 0.000, 0.000, 0.000]

    def test_raman2(self):
        assert self.data["raman2"] == [0.117, 10.597, 32.059, 0.895, 79.019, 71.489]

    def test_roa2(self):
        assert self.data["roa2"] == [0.000, -0.000, 0.000, 0.000, 0.000, 0.000]

    def test_cid2(self):
        assert self.data["cid2"] == [0.000, -0.000, 0.000, 0.000, 0.000, 0.000]

    def test_raman3(self):
        assert self.data["raman3"] == [0.233, 21.194, 64.118, 1.790, 158.038, 142.978]

    def test_roa3(self):
        assert self.data["roa3"] == [0.000, -0.000, -0.000, 0.000, 0.000, 0.000]

    def test_cid3(self):
        assert self.data["cid3"] == [0.000, -0.000, -0.000, 0.000, 0.000, 0.000]

    def test_rc180(self):
        assert self.data["rc180"] == [0.714, 0.714, 0.266, -0.398, -0.341, 0.714]


@pytest.mark.usefixtures("setup_class")
class TestOpt(ParserTestCase):
    file = "fal-opt.out"

    def test_methods_called(self):
        self.parser.initial.assert_called()
        self.parser.wait.assert_called()
        self.parser.frequencies.assert_not_called()
        self.parser.excited.assert_not_called()
        self.parser.optimization.assert_called()
        self.parser.geometry.assert_called()

    def test_keys(self):
        assert set(self.data.keys()) == {
            "normal_termination",
            "version",
            "command",
            "charge",
            "multiplicity",
            "input_geom",
            "input_atoms",
            "optimized_geom",
            "optimized_atoms",
            "scf",
            "stoichiometry",
            "last_read_geom",
            "last_read_atoms",
            "optimization_completed",
        }

    def test_optimization_completed(self):
        assert self.data["optimization_completed"]

    def test_command(self):
        assert self.data["command"] == "opt hf/3-21g"

    def test_scf(self):
        assert self.data["scf"] == -113.221819992

    def test_geometry(self):
        assert self.data["last_read_geom"] == [
            [-0.000016, 0.530659, 0.000000],
            [0.913392, 1.112945, 0.000000],
            [-0.913174, 1.113406, 0.000000],
            [-0.000016, -0.676288, 0.000000],
        ]

    def test_optimized(self):
        assert self.data["optimization_completed"]
        assert self.data["optimized_geom"] == [
            [-0.000016, 0.530659, 0.000000],
            [0.913392, 1.112945, 0.000000],
            [-0.913174, 1.113406, 0.000000],
            [-0.000016, -0.676288, 0.000000],
        ]


@pytest.mark.usefixtures("setup_class")
class TestOptFreq(ParserTestCase):
    file = "fal-opt-freq.out"

    def test_methods_called(self):
        self.parser.initial.assert_called()
        self.parser.wait.assert_called()
        self.parser.frequencies.assert_called()
        self.parser.excited.assert_not_called()
        self.parser.optimization.assert_called()
        self.parser.geometry.assert_called()

    def test_keys(self):
        assert set(self.data.keys()) == {
            "normal_termination",
            "version",
            "command",
            "charge",
            "multiplicity",
            "input_geom",
            "input_atoms",
            "optimized_geom",
            "optimized_atoms",
            "scf",
            "stoichiometry",
            "zpecorr",
            "tencorr",
            "entcorr",
            "gibcorr",
            "zpe",
            "ten",
            "ent",
            "gib",
            "freq",
            "mass",
            "frc",
            "iri",
            "depolarp",
            "depolaru",
            "ramanactiv",
            "last_read_geom",
            "last_read_atoms",
            "optimization_completed",
        }

    def test_command(self):
        assert self.data["command"] == "opt freq hf/3-21g"

    def test_geometry(self):
        assert self.data["last_read_geom"] == [
            [-0.000016, 0.530659, 0.000000],
            [0.913392, 1.112945, 0.000000],
            [-0.913174, 1.113406, 0.000000],
            [-0.000016, -0.676288, 0.000000],
        ]

    def test_optimized(self):
        assert self.data["optimization_completed"]
        assert self.data["optimized_geom"] == [
            [-0.000016, 0.530659, 0.000000],
            [0.913392, 1.112945, 0.000000],
            [-0.913174, 1.113406, 0.000000],
            [-0.000016, -0.676288, 0.000000],
        ]


@pytest.mark.usefixtures("setup_class")
class TestInputError(ParserTestCase):
    file = "fal-input-error.out"

    def test_keys(self):
        assert set(self.data.keys()) == {
            "normal_termination",
            "version",
            "command",
            "charge",
            "multiplicity",
            "input_geom",
            "input_atoms",
        }

    def test_normal_termination(self):
        assert not self.data["normal_termination"]

    def test_stoichiometry(self):
        assert "stoichiometry" not in self.data

    def test_atoms(self):
        assert "last_read_atoms" not in self.data


@pytest.mark.usefixtures("setup_class")
class TestTd(ParserTestCase):
    file = "fal-td.out"

    def test_methods_called(self):
        self.parser.initial.assert_called()
        self.parser.wait.assert_called()
        self.parser.frequencies.assert_not_called()
        self.parser.excited.assert_called()
        self.parser.optimization.assert_not_called()
        self.parser.geometry.assert_called()

    def test_keys(self):
        assert set(self.data.keys()) == {
            "normal_termination",
            "version",
            "command",
            "charge",
            "multiplicity",
            "input_geom",
            "input_atoms",
            "scf",
            "stoichiometry",
            "vdip",
            "vosc",
            "ldip",
            "losc",
            "vrot",
            "lrot",
            "eemang",
            "wavelen",
            "ex_en",
            "transitions",
            "last_read_geom",
            "last_read_atoms",
        }

    def test_command(self):
        assert self.data["command"] == "td hf/3-21g"

    def test_vdip(self):
        assert self.data["vdip"] == (0.0000, 0.0100, 0.0360)

    def test_vosc(self):
        assert self.data["vosc"] == (0.0000, 0.0218, 0.0714)

    def test_ldip(self):
        assert self.data["ldip"] == (0.0000, 0.0199, 0.8344)

    def test_losc(self):
        assert self.data["losc"] == (0.0000, 0.0040, 0.1869)

    def test_vrot(self):
        assert self.data["vrot"] == (0.0000, 0.0000, 0.0000)

    def test_lrot(self):
        assert self.data["lrot"] == (0.0000, 0.0000, 0.0000)

    def test_eemang(self):
        assert self.data["eemang"] == (90.00, 90.00, 90.00)

    def test_wavelen(self):
        assert self.data["wavelen"] == [326.42, 149.31, 135.60]

    def test_ex_en(self):
        assert self.data["ex_en"] == [3.7983, 8.3039, 9.1437]

    def test_transitions(self):
        assert self.data["transitions"] == [
            [[5, 9, 0.10410], [8, 9, 0.69982]],
            [
                [6, 9, 0.70461],
            ],
            [[6, 12, 0.12121], [7, 9, 0.68192], [8, 11, -0.11535]],
        ]


@pytest.mark.skip("To be created.")
@pytest.mark.usefixtures("setup_class")
class TestUnfinished(ParserTestCase):
    # TODO: create test suite for interrupted jobs
    file = ""

    def test_normal_termination(self):
        assert self.data["normal_termination"]
